import contextlib
import ctypes
import dataclasses
import logging
import os
import sys
import threading
import typing
from collections import namedtuple
from enum import Enum, auto
from functools import cached_property, wraps
from pathlib import Path
from select import select
from typing import Any, Optional

from xonsh.built_ins import XSH
from xonsh.main import main as xonsh_main

from rpcclient.clients.darwin.consts import BLOCK_IS_GLOBAL
from rpcclient.core.capture_fd import CaptureFD
from rpcclient.core.structs.consts import (
    EAGAIN,
    ECONNREFUSED,
    EEXIST,
    EISDIR,
    ENOENT,
    ENOTDIR,
    ENOTEMPTY,
    EPERM,
    EPIPE,
    RTLD_NEXT,
)
from rpcclient.core.structs.generic import block_descriptor, block_literal
from rpcclient.core.subsystems.decorator import subsystem
from rpcclient.core.subsystems.fs import Fs
from rpcclient.core.subsystems.lief import Lief
from rpcclient.core.subsystems.network import Network
from rpcclient.core.subsystems.processes import Processes
from rpcclient.core.subsystems.sysctl import Sysctl
from rpcclient.core.symbol import Symbol
from rpcclient.core.symbols_jar import SymbolsJar
from rpcclient.event_notifier import EventNotifier
from rpcclient.exceptions import (
    ArgumentError,
    BadReturnValueError,
    RpcBrokenPipeError,
    RpcConnectionRefusedError,
    RpcFileExistsError,
    RpcFileNotFoundError,
    RpcIsADirectoryError,
    RpcNotADirectoryError,
    RpcNotEmptyError,
    RpcPermissionError,
    RpcResourceTemporarilyUnavailableError,
    ServerResponseError,
    SpawnError,
)
from rpcclient.protocol.rpc_bridge import RpcBridge
from rpcclient.protos.rpc_api_pb2 import Argument, MsgId
from rpcclient.protos.rpc_pb2 import ProtocolConstants

tty_support = False
try:
    import termios
    import tty

    tty_support = True
except ImportError:
    logging.warning("termios not available on your system. some functionality may not work as expected")

io_or_str = typing.TypeVar("io_or_str", typing.IO, str)

SpawnResult = namedtuple("SpawnResult", "error pid stdout")

INVALID_PID = 0xFFFFFFFF
CHUNK_SIZE = 1024

USAGE = """
Welcome to the rpcclient interactive shell! You interactive shell for controlling the remote rpcserver.
Feel free to use the following globals:

🌍 p - the injected process
🌍 symbols - process global symbols

Have a nice flight ✈️!
Starting an IPython shell... 🐍
"""


@dataclasses.dataclass
class ProtocolDitentStat:
    errno: int
    st_dev: int  # device inode resides on
    st_mode: int  # inode protection mode
    st_nlink: int  # number of hard links to the file
    st_ino: int  # inode's number
    st_uid: int  # user-id of owner
    st_gid: int  # group-id of owner
    st_rdev: int  # device type, for special file inode
    st_size: int  # file size, in bytes
    st_blocks: int  # blocks allocated for file
    st_blksize: int  # optimal blocksize for I/O
    st_atime: int
    st_mtime: int
    st_ctime: int


@dataclasses.dataclass
class ProtocolDirent:
    d_inode: int
    d_type: int
    d_name: str
    lstat: ProtocolDitentStat
    stat: ProtocolDitentStat


class ClientEvent(Enum):
    CREATED = auto()
    TERMINATED = auto()


def null_pointer_guard(func: typing.Callable) -> typing.Callable:
    """
    A decorator to prevent dereferencing a null pointer by checking if the given
    address is zero before calling the wrapped function. If the address is zero,
    an ArgumentError is raised.

    Parameters:
        func (Callable): The function to be wrapped by the decorator.

    Returns:
        Callable: The wrapped function with null pointer guard applied.

    Raises:
        ArgumentError: If the 'address' argument is zero.
    """

    @wraps(func)
    def wrapper(self: typing.Any, address: int, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        if address == 0:
            raise ArgumentError("Unable dereference null pointer.")
        return func(self, address, *args, **kwargs)

    return wrapper


class CoreClient:
    """Main client interface to access the remote rpcserver"""

    DEFAULT_ARGV: typing.ClassVar = ["/bin/sh"]
    DEFAULT_ENVP: typing.ClassVar = []

    def __init__(self, bridge: RpcBridge, dlsym_global_handle=RTLD_NEXT):
        self._bridge = bridge
        self._old_settings = None
        self._endianness = "<"
        self._dlsym_global_handle = dlsym_global_handle
        self._protocol_lock = threading.Lock()
        self._logger = logging.getLogger(self.__module__)
        self.notifier = EventNotifier()
        self.symbols = SymbolsJar.create(self)

    @subsystem
    def fs(self) -> Fs:
        return Fs(self)

    @subsystem
    def processes(self) -> Processes:
        return Processes(self)

    @subsystem
    def network(self) -> Network:
        return Network(self)

    @subsystem
    def lief(self) -> Lief:
        return Lief(self)

    @subsystem
    def sysctl(self) -> Sysctl:
        return Sysctl(self)

    def info(self):
        """print information about the current target"""
        uname = self.uname
        print("sysname:", uname.sysname)
        print("nodename:", uname.nodename)
        print("release:", uname.release)
        print("version:", uname.version)
        print("machine:", uname.machine)
        print(f"uid: {self.symbols.getuid():d}")
        print(f"gid: {self.symbols.getgid():d}")
        print(f"pid: {self.pid:d}")
        print(f"ppid: {self.symbols.getppid():d}")
        print(f"progname: {self.progname}")

    @cached_property
    def progname(self) -> str:
        """get the program name from remote"""
        # bugfix https://github.com/doronz88/rpc-project/issues/405
        # Default Linux libraries don't expose `getprogname()`, so instead we use the `__progname` global symbol.
        # Tested on both macOS and Ubuntu.
        return self.symbols["__progname"][0].peek_str()

    @cached_property
    def uname(self):
        """get the utsname struct from remote"""
        raise NotImplementedError()

    @property
    def id(self) -> int:
        return self._bridge.client_id

    @property
    def platform(self) -> str:
        return self._bridge.platform

    @property
    def sysname(self) -> str:
        return self._bridge.sysname

    @property
    def arch(self) -> int:
        return self._bridge.arch

    def rpc_call(self, msg_id: int, **kwargs):
        try:
            return self._bridge.rpc_call(msg_id, **kwargs)
        except ConnectionError:
            self.notifier.notify(ClientEvent.TERMINATED, self.id)
            raise
        except ServerResponseError:
            raise

    def dlopen(self, filename: str, mode: int) -> Symbol:
        """Load a shared library on the remote host and return its handle."""
        return self.symbol(self.rpc_call(MsgId.REQ_DLOPEN, filename=filename, mode=mode).handle)

    def dlclose(self, lib: int) -> int:
        """Close a previously opened remote library handle."""
        return self.rpc_call(MsgId.REQ_DLCLOSE, handle=ctypes.c_uint64(lib).value).res

    def dlsym(self, lib: int, symbol_name: str) -> int:
        """Resolve a symbol name in a remote library handle and return its address."""
        return self.rpc_call(MsgId.REQ_DLSYM, handle=ctypes.c_uint64(lib).value, symbol_name=symbol_name).ptr

    @null_pointer_guard
    def call(
        self,
        address: int,
        argv: Optional[list[int]] = None,
        return_float64=False,
        return_float32=False,
        return_raw=False,
        va_list_index: int = 0xFFFF,
    ) -> typing.Union[float, Symbol, Any]:
        """call a remote function and retrieve its return value as a Symbol object"""
        args = []
        for arg in argv:
            if isinstance(arg, float):
                args.append(Argument(v_double=arg))
            elif isinstance(arg, str):
                args.append(Argument(v_str=arg))
            elif isinstance(arg, int):
                args.append(Argument(v_int=ctypes.c_uint64(arg).value))
            elif isinstance(arg, bytes):
                args.append(Argument(v_bytes=arg))
            elif isinstance(arg, Enum):
                args.append(Argument(v_int=ctypes.c_uint64(arg.value).value))
            else:
                raise ArgumentError()

        ret = self.rpc_call(MsgId.REQ_CALL, address=address, va_list_index=va_list_index, argv=args)
        if ret.HasField("arm_registers"):
            d0 = ret.arm_registers.d0
            if return_float32:
                return ctypes.c_float(d0).value
            if return_float64:
                return d0
            if return_raw:
                return ret.arm_registers
            return self.symbol(ret.arm_registers.x0)
        return self.symbol(ret.return_value)

    @null_pointer_guard
    def peek(self, address: int, size: int) -> bytes:
        """peek data at the given address"""
        try:
            return self.rpc_call(MsgId.REQ_PEEK, address=address, size=size).data
        except ServerResponseError as e:
            raise ArgumentError() from e

    @null_pointer_guard
    def poke(self, address: int, data: bytes):
        """poke data at a given address"""
        try:
            return self.rpc_call(MsgId.REQ_POKE, address=address, data=data)
        except ServerResponseError as e:
            raise ArgumentError() from e

    def get_dummy_block(self) -> Symbol:
        """Get an address for a stub block containing nothing"""
        block_size = block_literal.sizeof()
        desc_size = block_descriptor.sizeof()

        descriptor = self.symbols.malloc(desc_size)
        descriptor.poke(block_descriptor.build({"reserved": 0, "size": block_size}))

        block = self.symbols.malloc(block_size)
        block.poke(
            block_literal.build({
                "isa": self.symbols._NSConcreteGlobalBlock,
                "flags": BLOCK_IS_GLOBAL,
                "reserved": 0,
                "invoke": self.symbols.getpid,
                "descriptor": descriptor,
            })
        )
        return block

    def listdir(self, path: str):
        """get an address for a stub block containing nothing"""
        entries = []
        try:
            ret = self.rpc_call(MsgId.REQ_LIST_DIR, path=path)
        except ServerResponseError:
            self.raise_errno_exception(f"failed to listdir: {path}")

        for entry in ret.dir_entries:
            lstat = ProtocolDitentStat(
                errno=entry.lstat.errno1,
                st_blocks=entry.lstat.st_blocks,
                st_blksize=entry.lstat.st_blksize,
                st_atime=entry.lstat.st_atime1,
                st_ctime=entry.lstat.st_ctime1,
                st_mtime=entry.lstat.st_mtime1,
                st_nlink=entry.lstat.st_nlink,
                st_mode=entry.lstat.st_mode,
                st_rdev=entry.lstat.st_rdev,
                st_size=entry.lstat.st_size,
                st_dev=entry.lstat.st_dev,
                st_gid=entry.lstat.st_gid,
                st_ino=entry.lstat.st_ino,
                st_uid=entry.lstat.st_uid,
            )
            stat = ProtocolDitentStat(
                errno=entry.stat.errno1,
                st_blocks=entry.stat.st_blocks,
                st_blksize=entry.stat.st_blksize,
                st_atime=entry.stat.st_atime1,
                st_ctime=entry.stat.st_ctime1,
                st_mtime=entry.stat.st_mtime1,
                st_nlink=entry.stat.st_nlink,
                st_mode=entry.stat.st_mode,
                st_rdev=entry.stat.st_rdev,
                st_size=entry.stat.st_size,
                st_dev=entry.stat.st_dev,
                st_gid=entry.stat.st_gid,
                st_ino=entry.stat.st_ino,
                st_uid=entry.stat.st_uid,
            )
            entries.append(
                ProtocolDirent(
                    d_inode=entry.lstat.st_ino, d_type=entry.d_type, d_name=entry.d_name, lstat=lstat, stat=stat
                )
            )
        return entries

    def spawn(
        self,
        argv: Optional[list[str]] = None,
        envp: Optional[list[str]] = None,
        stdin: io_or_str = sys.stdin,
        stdout=sys.stdout,
        raw_tty=False,
        background=False,
    ) -> SpawnResult:
        """
        spawn a new process and forward its stdin, stdout & stderr

        :param argv: argv of the process to be executed
        :param envp: envp of the process to be executed
        :param stdin: either a file object to read from OR a string
        :param stdout: a file object to write both stdout and stderr to. None if background is requested
        :param raw_tty: should enable raw tty mode
        :param background: should execute process in background
        :return: a SpawnResult. error is None if background is requested
        """
        if argv is None:
            argv = self.DEFAULT_ARGV

        if envp is None:
            envp = self.DEFAULT_ENVP

        with self._protocol_lock:
            pid = self._execute(argv, envp, background=background)
            self._logger.info(f"shell process started as pid: {pid}")

            if background:
                return SpawnResult(error=None, pid=pid, stdout=None)

            if raw_tty:
                self._prepare_terminal()
            try:
                error = self.enter_pty_mode(stdin, stdout)
            except Exception:
                # this is important to really catch every exception here, even exceptions not inheriting from Exception
                # so the controlling terminal will remain working with its previous settings
                if raw_tty:
                    self._restore_terminal()
                raise

        return SpawnResult(error=error, pid=pid, stdout=stdout)

    def symbol(self, symbol: int):
        """at a symbol object from a given address"""
        return Symbol.create(symbol, self)

    @property
    def errno(self) -> int:
        return self.symbols.errno[0]

    @errno.setter
    def errno(self, value: int) -> None:
        self.symbols.errno[0] = value

    @property
    def last_error(self):
        """get info about the last occurred error"""
        if not self.errno:
            return ""
        err_str = self.symbols.strerror(self.errno).peek_str()
        return f"[{self.errno}] {err_str}"

    @property
    def environ(self) -> list[str]:
        result = []
        environ = self.symbols.environ[0]
        i = 0
        while environ[i]:
            result.append(environ[i].peek_str())
            i += 1
        return result

    def setenv(self, name: str, value: str):
        """set process environment variable"""
        self.symbols.setenv(name, value)

    def getenv(self, name: str) -> typing.Optional[str]:
        """get process environment variable"""
        value = self.symbols.getenv(name)
        if not value:
            return None
        return value.peek_str()

    @cached_property
    def pid(self):
        return int(self.symbols.getpid())

    @contextlib.contextmanager
    def safe_calloc(self, size: int):
        with self.safe_malloc(size) as x:
            x.poke(b"\x00" * size)
            try:
                yield x
            finally:
                pass

    @contextlib.contextmanager
    def safe_malloc(self, size: int):
        with self.freeing(self.symbols.malloc(size)) as x:
            yield x

    @contextlib.contextmanager
    def freeing(self, symbol):
        try:
            yield symbol
        finally:
            if symbol:
                self.symbols.free(symbol)

    def close(self):
        self.rpc_call(MsgId.REQ_CLOSE_CLIENT)
        self.notifier.notify(ClientEvent.TERMINATED, self.id)
        self._bridge.close()

    def shell(self):
        self._logger.disabled = True

        args = ["--rc"]
        args.append(str((Path(__file__).parent / "xonshrc.py").absolute()))

        XSH.ctx["_client_to_reuse"] = self

        try:
            logging.getLogger("parso.python.diff").disabled = True
            logging.getLogger("parso.cache").disabled = True
            logging.getLogger("asyncio").disabled = True
            xonsh_main(args)
        except SystemExit:
            self._logger.disabled = False

    def _execute(self, argv: list[str], envp: list[str], background=False) -> int:
        try:
            return self.rpc_call(MsgId.REQ_EXEC, background=background, argv=argv, envp=envp).pid
        except ServerResponseError as e:
            raise SpawnError(f"failed to spawn: {argv}") from e

    def enter_pty_mode(self, stdin=sys.stdin, stdout=sys.stdout):
        # the socket must be non-blocking for using select()
        sock = self._bridge.sock.raw_socket
        sock.setblocking(False)
        exit_code = None
        try:
            fds = []
            if hasattr(stdin, "fileno"):
                fds.append(stdin)
            else:
                data = stdin
                if isinstance(data, str):
                    data = data.encode()
                sock.sendall(data)
            fds.append(sock)

            running = True
            while running:
                rlist, _, _ = select(fds, [], [])
                for fd in rlist:
                    if fd == stdin or (stdin is sys.stdin and fd == sys.stdin):
                        if stdin is sys.stdin:
                            buf = os.read(stdin.fileno(), ProtocolConstants.RPC_PTY_BUFFER_SIZE)
                        else:
                            buf = stdin.read(ProtocolConstants.RPC_PTY_BUFFER_SIZE)
                            if isinstance(buf, str):
                                buf = buf.encode()
                        if buf:
                            sock.sendall(buf)
                    elif fd == sock:
                        try:
                            response = self._bridge.sock.rpc_msg_recv_pty()
                        except ConnectionResetError:
                            print("Bye. 👋")
                            running = False
                            break
                        msg_type = response.WhichOneof("type")
                        if msg_type == "buffer":
                            stdout.write(response.buffer.decode())
                            if hasattr(stdout, "flush"):
                                stdout.flush()
                        elif msg_type == "exit_code":
                            exit_code = response.exit_code
                            running = False
                            break
        finally:
            sock.setblocking(True)
        return exit_code

    def _restore_terminal(self):
        if not tty_support:
            return
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, self._old_settings)

    def _prepare_terminal(self):
        if not tty_support:
            return
        fd = sys.stdin
        self._old_settings = termios.tcgetattr(fd)
        tty.setraw(fd)

    def raise_errno_exception(self, message: str):
        message += f" ({self.last_error})"
        exceptions = {
            EPERM: RpcPermissionError,
            ENOENT: RpcFileNotFoundError,
            EEXIST: RpcFileExistsError,
            EISDIR: RpcIsADirectoryError,
            ENOTDIR: RpcNotADirectoryError,
            EPIPE: RpcBrokenPipeError,
            ENOTEMPTY: RpcNotEmptyError,
            EAGAIN: RpcResourceTemporarilyUnavailableError,
            ECONNREFUSED: RpcConnectionRefusedError,
        }
        exception = exceptions.get(self.errno)
        if exception:
            raise exception(message)
        raise BadReturnValueError(message)

    def capture_fd(self, fd: int, sock_buf_size: Optional[int] = None) -> CaptureFD:
        """
        Get a context manager, capturing output to `fd`. Read from it using the `read()` method

        sock_buf_size is required for captures above 6KB, as any write above this value would block until a read is performed.

        :param fd: FD to capture
        :param sock_buf_size: Buffer size for the capture socket, if not specified, default value is used.
        :return: CaptureFD object
        """
        return CaptureFD(self, fd, sock_buf_size)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.pid} | {self.progname}>"



from construct import (
    Adapter,
    Aligned,
    Array,
    BitsInteger,
    BitStruct,
    Bytes,
    Computed,
    CString,
    Default,
    Enum,
    FlagsEnum,
    GreedyRange,
    Hex,
    Int8ul,
    Int16sl,
    Int16ub,
    Int16ul,
    Int32sl,
    Int32ub,
    Int32ul,
    Int64sl,
    Int64ul,
    LazyArray,
    Octet,
    PaddedString,
    Padding,
    Seek,
    Struct,
    Switch,
    Tell,
    Union,
    this,
)

from rpcclient.core.structs.consts import AF_INET, AF_INET6, AF_UNIX
from rpcclient.core.structs.generic import (
    UNIX_PATH_MAX,
    gid_t,
    in_addr,
    long,
    mode_t,
    short,
    st_flags,
    time_t,
    u_char,
    u_int32_t,
    u_short,
    uid_t,
    uint8_t,
    uint32_t,
    uint64_t,
)

MAXPATHLEN = 1024
_SYS_NAMELEN = 256

utsname = Struct(
    "sysname" / PaddedString(_SYS_NAMELEN, "utf8"),
    "nodename" / PaddedString(_SYS_NAMELEN, "utf8"),
    "release" / PaddedString(_SYS_NAMELEN, "utf8"),
    "version" / PaddedString(_SYS_NAMELEN, "utf8"),
    "machine" / PaddedString(_SYS_NAMELEN, "utf8"),
)

pid_t = Int32ul
exitcode_t = Int32sl
ino_t = Int32ul
dev_t = Int32ul
off_t = Int64ul
nlink_t = Int16ul
blkcnt_t = Int64ul
blksize_t = Int32ul
ino64_t = Int64ul
fsid_t = Array(2, Int32sl)
io_name_t = PaddedString(1024, "utf8")
io_object_t = Int64ul
vm_prot_t = Int32ul
vm_inherit_t = Int32ul
boolean_t = Int32ul
memory_object_offset_t = Int64ul
vm_behavior_t = Int32ul
mach_vm_address_t = Int64ul
mach_vm_size_t = Int64ul
integer_t = Int32sl
natural_t = Int32ul
mach_port_name_t = natural_t
integer_t = Int32sl
mach_port_t = natural_t
mach_port_type_t = natural_t
mach_port_urefs_t = natural_t

timespec = Struct(
    "tv_sec" / long,
    "tv_nsec" / long,
)
dirent32 = Struct(
    "d_ino" / ino_t,
    "d_offset" / Int16ul,
    "d_reclen" / Int8ul,
    "d_namelen" / Int8ul,
    "d_name" / PaddedString(this.d_namelen, "utf8"),
)
dirent64 = Struct(
    "d_ino" / ino64_t,
    "d_offset" / Int64ul,
    "d_reclen" / Int16ul,
    "d_namelen" / Int16ul,
    "d_type" / Int8ul,
    "d_name" / PaddedString(this.d_namelen, "utf8"),
)
stat32 = Struct(
    "st_dev" / dev_t,  # device inode resides on
    "st_ino" / ino_t,  # inode's number
    "st_mode" / mode_t,  # inode protection mode
    "st_nlink" / nlink_t,  # number of hard links to the file
    "st_uid" / uid_t,  # user-id of owner
    "st_gid" / gid_t,  # group-id of owner
    "st_rdev" / dev_t,  # device type, for special file inode
    "st_atimespec" / timespec,  # time of last access
    "st_atime" / Computed(this.st_atimespec.tv_sec + (this.st_atimespec.tv_nsec / 10**9)),
    "st_mtimespec" / timespec,  # time of last data modification
    "st_mtime" / Computed(this.st_mtimespec.tv_sec + (this.st_mtimespec.tv_nsec / 10**9)),
    "st_ctimespec" / timespec,  # time of last file status change
    "st_ctime" / Computed(this.st_ctimespec.tv_sec + (this.st_ctimespec.tv_nsec / 10**9)),
    "st_size" / off_t,  # file size, in bytes
    "st_blocks" / blkcnt_t,  # blocks allocated for file
    "st_blksize" / blksize_t,  # optimal blocksize for I/O
    "st_flags" / Int32ul,  # user defined flags for file
    "st_gen" / Int32ul,
    "st_lspare" / Int32sl,
    "st_qspare" / Array(2, Int64sl),
    Padding(24),  # Compiled size is 144
)
stat64 = Struct(
    "st_dev" / dev_t,  # device inode resides on
    "st_mode" / mode_t,  # inode protection mode
    "st_nlink" / nlink_t,  # number of hard links to the file
    "st_ino" / ino64_t,  # inode's number
    "st_uid" / uid_t,  # user-id of owner
    "st_gid" / gid_t,  # group-id of owner
    "st_rdev" / dev_t,  # device type, for special file inode
    Padding(4),
    "st_atimespec" / timespec,  # time of last access
    "st_atime" / Computed(this.st_atimespec.tv_sec + (this.st_atimespec.tv_nsec / 10**9)),
    "st_mtimespec" / timespec,  # time of last data modification
    "st_mtime" / Computed(this.st_mtimespec.tv_sec + (this.st_mtimespec.tv_nsec / 10**9)),
    "st_ctimespec" / timespec,  # time of last file status change
    "st_ctime" / Computed(this.st_ctimespec.tv_sec + (this.st_ctimespec.tv_nsec / 10**9)),
    "st_birthtimespec" / timespec,  # time of file creation(birth)
    "st_btime" / Computed(this.st_birthtimespec.tv_sec + (this.st_birthtimespec.tv_nsec / 10**9)),
    "st_size" / off_t,
    "st_blocks" / blkcnt_t,  # blocks allocated for file
    "st_blksize" / blksize_t,  # optimal blocksize for I/O
    "st_flags" / st_flags,  # blocks allocated for file
    "st_gen" / Int32ul,  # user defined flags for file
    # seems like this value doesn't really exist
    # 'st_lspare' / Int32ul,
    "st_qspare" / Array(2, Int64sl),
)

MFSNAMELEN = 15  # length of fs type name, not inc. nul
MNAMELEN = 90  # length of buffer for returned name

# when _DARWIN_FEATURE_64_BIT_INODE is NOT defined
statfs64 = Struct(
    "f_otype" / Int16sl,  # type of file system (reserved: zero)
    "f_oflags" / Int16sl,  # copy of mount flags (reserved: zero)
    Padding(4),
    "f_bsize" / Int64ul,  # fundamental file system block size
    "f_iosize" / Int64ul,  # optimal transfer block size
    "f_blocks" / Int64ul,  # total data blocks in file system
    "f_bfree" / Int64ul,  # free blocks in fs
    "f_bavail" / Int64ul,  # free blocks avail to non-superuser
    "f_files" / Int64ul,  # total file nodes in file system
    "f_ffree" / Int64ul,  # free file nodes in fs
    "f_fsid" / fsid_t,  # file system id
    "f_owner" / uid_t,  # user that mounted the file system
    Padding(4),
    "f_reserved1" / Int16sl,  # reserved for future use
    "f_type" / Int16sl,  # type of file system (reserved)
    Padding(4),
    "f_flags" / Int64ul,  # copy of mount flags
    "f_reserved2" / Int64ul,  # reserved for future use
    "f_fstypename" / PaddedString(MFSNAMELEN, "utf8"),  # fs type name
    "f_mntonname" / PaddedString(MNAMELEN, "utf8"),  # directory on which mounted
    "f_mntfromname" / PaddedString(MNAMELEN, "utf8"),  # mounted file system
    "f_reserved3" / Int8ul,  # reserved for future use
    "f_reserved4" / Int64ul,  # reserved for future use
)

MFSTYPENAMELEN = 16  # length of fs type name including null
MNAMELEN = MAXPATHLEN

# when _DARWIN_FEATURE_64_BIT_INODE is defined
statfs32 = Struct(
    "f_bsize" / Int32ul,  # fundamental file system block size
    "f_iosize" / Int32sl,  # optimal transfer block size
    "f_blocks" / Int64ul,  # total data blocks in file system
    "f_bfree" / Int64ul,  # free blocks in fs
    "f_bavail" / Int64ul,  # free blocks avail to non-superuser
    "f_files" / Int64ul,  # total file nodes in file system
    "f_ffree" / Int64ul,  # free file nodes in fs
    "f_fsid" / fsid_t,  # file system id
    "f_owner" / uid_t,  # user that mounted the filesystem
    "f_type" / Int32ul,  # type of filesystem
    "f_flags" / Int32ul,  # copy of mount exported flags
    "f_fssubtype" / Int32ul,  # fs sub-type (flavor)
    "f_fstypename" / PaddedString(MFSTYPENAMELEN, "utf8"),  # fs type name
    "f_mntonname" / PaddedString(MAXPATHLEN, "utf8"),  # directory on which mounted
    "f_mntfromname" / PaddedString(MAXPATHLEN, "utf8"),  # mounted filesystem
    "f_reserved" / Int32ul[8],  # For future use
)

PROC_ALL_PIDS = 1

# Flavors for proc_pidinfo()
PROC_PIDLISTFDS = 1
PROC_PIDTASKALLINFO = 2
PROC_PIDTBSDINFO = 3
PROC_PIDTASKINFO = 4
PROC_PIDTHREADINFO = 5
PROC_PIDLISTTHREADS = 6
PROC_PIDREGIONINFO = 7
PROC_PIDREGIONPATHINFO = 8
PROC_PIDVNODEPATHINFO = 9
PROC_PIDTHREADPATHINFO = 10
PROC_PIDPATHINFO = 11
PROC_PIDWORKQUEUEINFO = 12
PROC_PIDT_SHORTBSDINFO = 13
PROC_PIDLISTFILEPORTS = 14
PROC_PIDTHREADID64INFO = 15
PROC_PID_RUSAGE = 16
PROC_PIDUNIQIDENTIFIERINFO = 17
PROC_PIDT_BSDINFOWITHUNIQID = 18
PROC_PIDARCHINFO = 19
PROC_PIDCOALITIONINFO = 20
PROC_PIDNOTEEXIT = 21
PROC_PIDREGIONPATHINFO2 = 22
PROC_PIDREGIONPATHINFO3 = 23

# Flavors for proc_pidfdinfo
PROC_PIDFDVNODEINFO = 1
PROC_PIDFDVNODEPATHINFO = 2
PROC_PIDFDSOCKETINFO = 3
PROC_PIDFDPSEMINFO = 4
PROC_PIDFDPSHMINFO = 5
PROC_PIDFDPIPEINFO = 6
PROC_PIDFDKQUEUEINFO = 7
PROC_PIDFDATALKINFO = 8

# Flavors for proc_pidfileportinfo
PROC_PIDFILEPORTVNODEPATHINFO = 2  # out: vnode_fdinfowithpath
PROC_PIDFILEPORTPSHMINFO = 5  # out: pshm_fdinfo
PROC_PIDFILEPORTPIPEINFO = 6  # out: pipe_fdinfo
# used for proc_setcontrol
PROC_SELFSET_PCONTROL = 1
PROC_SELFSET_THREADNAME = 2
PROC_SELFSET_VMRSRCOWNER = 3
PROC_SELFSET_DELAYIDLESLEEP = 4
# used for proc_dirtycontrol
PROC_DIRTYCONTROL_TRACK = 1
PROC_DIRTYCONTROL_SET = 2
PROC_DIRTYCONTROL_GET = 3
PROC_DIRTYCONTROL_CLEAR = 4
# proc_track_dirty() flags
PROC_DIRTY_TRACK = 0x1
PROC_DIRTY_ALLOW_IDLE_EXIT = 0x2
PROC_DIRTY_DEFER = 0x4
PROC_DIRTY_LAUNCH_IN_PROGRESS = 0x8
# proc_get_dirty() flags
PROC_DIRTY_TRACKED = 0x1
PROC_DIRTY_ALLOWS_IDLE_EXIT = 0x2
PROC_DIRTY_IS_DIRTY = 0x4
PROC_DIRTY_LAUNCH_IS_IN_PROGRESS = 0x8

# Flavors for proc_pidoriginatorinfo
PROC_PIDORIGINATOR_UUID = 0x1
PROC_PIDORIGINATOR_BGSTATE = 0x2
# __proc_info() call numbers
PROC_INFO_CALL_LISTPIDS = 0x1
PROC_INFO_CALL_PIDINFO = 0x2
PROC_INFO_CALL_PIDFDINFO = 0x3
PROC_INFO_CALL_KERNMSGBUF = 0x4
PROC_INFO_CALL_SETCONTROL = 0x5
PROC_INFO_CALL_PIDFILEPORTINFO = 0x6
PROC_INFO_CALL_TERMINATE = 0x7
PROC_INFO_CALL_DIRTYCONTROL = 0x8
PROC_INFO_CALL_PIDRUSAGE = 0x9
PROC_INFO_CALL_PIDORIGINATORINFO = 0xA

# defns of process file desc type
PROX_FDTYPE_ATALK = 0
PROX_FDTYPE_VNODE = 1
PROX_FDTYPE_SOCKET = 2
PROX_FDTYPE_PSHM = 3
PROX_FDTYPE_PSEM = 4
PROX_FDTYPE_KQUEUE = 5
PROX_FDTYPE_PIPE = 6
PROX_FDTYPE_FSEVENTS = 7

pbi_flags_t = FlagsEnum(
    Int32ul,
    # pbi_flags values
    PROC_FLAG_SYSTEM=1,  # System process
    PROC_FLAG_TRACED=2,  # process currently being traced, possibly by gdb
    PROC_FLAG_INEXIT=4,  # process is working its way in exit()
    PROC_FLAG_PPWAIT=8,
    PROC_FLAG_LP64=0x10,  # 64bit process
    PROC_FLAG_SLEADER=0x20,  # The process is the session leader
    PROC_FLAG_CTTY=0x40,  # process has a control tty
    PROC_FLAG_CONTROLT=0x80,  # Has a controlling terminal
    PROC_FLAG_THCWD=0x100,  # process has a thread with cwd
    # process control bits for resource starvation
    PROC_FLAG_PC_THROTTLE=0x200,
    # In resource starvation situations, this process is to be throttled
    PROC_FLAG_PC_SUSP=0x400,  # In resource starvation situations, this process is to be suspended
    PROC_FLAG_PC_KILL=0x600,  # In resource starvation situations, this process is to be terminated
    PROC_FLAG_PC_MASK=0x600,
    # process action bits for resource starvation
    PROC_FLAG_PA_THROTTLE=0x800,  # The process is currently throttled due to resource starvation
    PROC_FLAG_PA_SUSP=0x1000,  # The process is currently suspended due to resource starvation
    PROC_FLAG_PSUGID=0x2000,  # process has set privileges since last exec
    PROC_FLAG_EXEC=0x4000,  # process has called exec
    PROC_FLAG_DARWINBG=0x8000,  # process in darwin background
    PROC_FLAG_EXT_DARWINBG=0x10000,  # process in darwin background - external enforcement
    PROC_FLAG_IOS_APPLEDAEMON=0x20000,  # Process is apple daemon
    PROC_FLAG_DELAYIDLESLEEP=0x40000,  # Process is marked to delay idle sleep on disk IO
    PROC_FLAG_IOS_IMPPROMOTION=0x80000,  # Process is daemon which receives importane donation
    PROC_FLAG_ADAPTIVE=0x100000,  # Process is adaptive
    PROC_FLAG_ADAPTIVE_IMPORTANT=0x200000,  # Process is adaptive, and is currently important
    PROC_FLAG_IMPORTANCE_DONOR=0x400000,  # Process is marked as an importance donor
    PROC_FLAG_SUPPRESSED=0x800000,  # Process is suppressed
    PROC_FLAG_APPLICATION=0x1000000,  # Process is an application
    PROC_FLAG_IOS_APPLICATION=0x1000000,  # Process is an application
)

MAXCOMLEN = 16

proc_bsdinfo = Struct(
    "pbi_flags" / pbi_flags_t,  # 64bit emulated etc
    "pbi_status" / Int32ul,
    "pbi_xstatus" / Int32ul,
    "pbi_pid" / Int32ul,
    "pbi_ppid" / Int32ul,
    "pbi_uid" / uid_t,
    "pbi_gid" / gid_t,
    "pbi_ruid" / uid_t,
    "pbi_rgid" / gid_t,
    "pbi_svuid" / uid_t,
    "pbi_svgid" / gid_t,
    "rfu_1" / Int32ul,  # reserved
    "_pbi_comm" / Bytes(MAXCOMLEN),
    "pbi_comm" / Computed(lambda x: x._pbi_comm.split(b"\x00", 1)[0].decode()),
    "_pbi_name" / Bytes(2 * MAXCOMLEN),  # empty if no name is registered
    "pbi_name" / Computed(lambda x: x._pbi_name.split(b"\x00", 1)[0].decode()),
    "pbi_nfiles" / Int32ul,
    "pbi_pgid" / Int32ul,
    "pbi_pjobc" / Int32ul,
    "e_tdev" / Int32ul,  # controlling tty dev
    "e_tpgid" / Int32ul,  # tty process group id
    "pbi_nice" / Int32sl,
    "pbi_start_tvsec" / Int64ul,
    "pbi_start_tvusec" / Int64ul,
)

proc_taskinfo = Struct(
    "pti_virtual_size" / Int64ul,  # virtual memory size (bytes)
    "pti_resident_size" / Int64ul,  # resident memory size (bytes)
    "pti_total_user" / Int64ul,  # total time
    "pti_total_system" / Int64ul,
    "pti_threads_user" / Int64ul,  # existing threads only
    "pti_threads_system" / Int64ul,
    "pti_policy" / Int32sl,  # default policy for new threads
    "pti_faults" / Int32sl,  # number of page faults
    "pti_pageins" / Int32sl,  # number of actual pageins
    "pti_cow_faults" / Int32sl,  # number of copy-on-write faults
    "pti_messages_sent" / Int32sl,  # number of messages sent
    "pti_messages_received" / Int32sl,  # number of messages received
    "pti_syscalls_mach" / Int32sl,  # number of mach system calls
    "pti_syscalls_unix" / Int32sl,  # number of unix system calls
    "pti_csw" / Int32sl,  # number of context switches
    "pti_threadnum" / Int32sl,  # number of threads in the task
    "pti_numrunning" / Int32sl,  # number of running threads
    "pti_priority" / Int32sl,  # task priority
)

proc_taskallinfo = Struct(
    "pbsd" / proc_bsdinfo,
    "ptinfo" / proc_taskinfo,
)

proc_regioninfo = Struct(
    "pri_protection" / Int32ul,
    "pri_max_protection" / Int32ul,
    "pri_inheritance" / Int32ul,
    "pri_flags" / Int32ul,  # shared, external pager, is submap
    "pri_offset" / Int64ul,
    "pri_behavior" / Int32ul,
    "pri_user_wired_count" / Int32ul,
    "pri_user_tag" / Int32ul,
    "pri_pages_resident" / Int32ul,
    "pri_pages_shared_now_private" / Int32ul,
    "pri_pages_swapped_out" / Int32ul,
    "pri_pages_dirtied" / Int32ul,
    "pri_ref_count" / Int32ul,
    "pri_shadow_depth" / Int32ul,
    "pri_share_mode" / Int32ul,
    "pri_private_pages_resident" / Int32ul,
    "pri_shared_pages_resident" / Int32ul,
    "pri_obj_id" / Int32ul,
    "pri_depth" / Int32ul,
    "pri_address" / Int64ul,
    "pri_size" / Int64ul,
)

proc_fdinfo = Struct(
    "proc_fd" / Int32sl,
    "proc_fdtype" / Int32ul,
)

proc_fileinfo = Struct(
    "fi_openflags" / Int32ul,
    "fi_status" / Int32ul,
    "fi_offset" / off_t,
    "fi_guardflags" / Int32ul,
)

# A copy of stat64 with static sized fields.
vinfo_stat = stat64

vnode_info = Struct(
    "vi_stat" / vinfo_stat,
    "vi_type" / Int32sl,
    "vi_pad" / Int32sl,
    "vi_fsid" / fsid_t,
)

vnode_info_path = Struct(
    "vip_vi" / vnode_info,
    "_vip_path" / Bytes(MAXPATHLEN),
    "vip_path" / Computed(lambda x: x._vip_path.split(b"\x00", 1)[0].decode()),
)

vnode_fdinfowithpath = Struct(
    "pfi" / proc_fileinfo,
    "pvip" / vnode_info_path,
)

sockbuf_info = Struct(
    "sbi_cc" / uint32_t,
    "sbi_hiwat" / uint32_t,  # SO_RCVBUF, SO_SNDBUF
    "sbi_mbcnt" / uint32_t,
    "sbi_mbmax" / uint32_t,
    "sbi_lowat" / uint32_t,
    "sbi_flags" / short,
    "sbi_timeo" / short,
)

# TCP Sockets

TSI_T_REXMT = 0  # retransmit
TSI_T_PERSIST = 1  # retransmit persistence
TSI_T_KEEP = 2  # keep alive
TSI_T_2MSL = 3  # 2*msl quiet time timer
TSI_T_NTIMERS = 4


class IpAddressAdapter(Adapter):
    def _decode(self, obj, context, path):
        return ".".join(map(str, obj))

    def _encode(self, obj, context, path):
        return list(map(int, obj.split(".")))


in4in6_addr = Struct(
    "i46a_pad32" / u_int32_t[3],
    "i46a_addr4" / IpAddressAdapter(in_addr),
)

in6_addr = Struct(
    "i46a_pad32" / u_int32_t[3],
    "i46a_addr4" / IpAddressAdapter(in_addr),
)

in_sockinfo = Struct(
    "insi_fport" / Int16ub,
    Padding(2),
    "insi_lport" / Int16ub,
    Padding(2),
    "insi_gencnt" / uint64_t,
    "insi_flags" / uint32_t,
    "insi_flow" / uint32_t,
    "insi_vflag" / uint8_t,
    "insi_ip_ttl" / uint8_t,
    Padding(2),
    "rfu_1" / uint32_t,
    # protocol dependent part
    "insi_faddr" / Union(None, "ina_46" / in4in6_addr, "ina_6" / in6_addr),
    "insi_laddr" / Union(None, "ina_46" / in4in6_addr, "ina_6" / in6_addr),
    "insi_v4"
    / Struct(
        "in4_tos" / u_char,
    ),
    "insi_v6"
    / Struct(
        "in6_hlim" / uint8_t,
        "in6_cksum" / Int32sl,
        "in6_ifindex" / u_short,
        "in6_hops" / short,
    ),
)

tcp_sockinfo = Struct(
    "tcpsi_ini" / in_sockinfo,
    "in_sockinfo" / Int32sl,
    "tcpsi_timer" / Int32sl[TSI_T_NTIMERS],
    "tcpsi_mss" / Int32sl,
    "tcpsi_flags" / uint32_t,
    "rfu_1" / uint32_t,  # reserved
    "tcpsi_tp" / uint64_t,  # opaque handle of TCP protocol control block
)

SOCK_MAXADDRLEN = 255

# Unix Domain Sockets

# we can't use sockaddr_un since the sun_path may contain utf8 invalid characters
sockaddr_un_raw = Struct(
    "sun_len" / Int8ul,
    "sun_family" / Default(Int8ul, AF_UNIX),
    "_sun_path" / Bytes(UNIX_PATH_MAX),
    "sun_path" / Computed(lambda x: x._sun_path.split(b"\x00", 1)[0].decode()),
)

un_sockinfo = Struct(
    "unsi_conn_so" / uint64_t,
    "unsi_conn_pcb" / uint64_t,
    "unsi_addr" / Union(None, "ua_sun" / sockaddr_un_raw, "ua_dummy" / Bytes(SOCK_MAXADDRLEN)),
    "unsi_caddr" / Union(None, "ua_sun" / sockaddr_un_raw, "ua_dummy" / Bytes(SOCK_MAXADDRLEN)),
)

IF_NAMESIZE = 16

# PF_NDRV Sockets

ndrv_info = Struct(
    "ndrvsi_if_family" / uint32_t,
    "ndrvsi_if_unit" / uint32_t,
    "ndrvsi_if_name" / Bytes(IF_NAMESIZE),
)

# Kernel Event Sockets

kern_event_info = Struct(
    "kesi_vendor_code_filter" / uint32_t,
    "kesi_class_filter" / uint32_t,
    "kesi_subclass_filter" / uint32_t,
)

# Kernel Control Sockets

MAX_KCTL_NAME = 96

kern_ctl_info = Struct(
    "kcsi_id" / uint32_t,
    "kcsi_reg_unit" / uint32_t,
    "kcsi_flags" / uint32_t,
    "kcsi_recvbufsize" / uint32_t,
    "kcsi_sendbufsize" / uint32_t,
    "kcsi_unit" / uint32_t,
    "kcsi_name" / Bytes(MAX_KCTL_NAME),
)

so_kind_t = Enum(
    Int32ul,
    SOCKINFO_GENERIC=0,
    SOCKINFO_IN=1,
    SOCKINFO_TCP=2,
    SOCKINFO_UN=3,
    SOCKINFO_NDRV=4,
    SOCKINFO_KERN_EVENT=5,
    SOCKINFO_KERN_CTL=6,
)

so_family_t = Enum(
    Int32ul,
    AF_INET=AF_INET,
    AF_INET6=AF_INET6,
)

socket_info = Struct(
    "soi_stat" / vinfo_stat,
    "soi_so" / uint64_t,  # opaque handle of socket
    "soi_pcb" / uint64_t,  # opaque handle of protocol control block
    "soi_type" / Int32sl,
    "soi_protocol" / Int32sl,
    "soi_family" / so_family_t,
    "soi_options" / short,
    "soi_linger" / short,
    "soi_state" / short,
    "soi_qlen" / short,
    "soi_incqlen" / short,
    "soi_qlimit" / short,
    "soi_timeo" / short,
    "soi_error" / u_short,
    "soi_oobmark" / uint32_t,
    "soi_rcv" / sockbuf_info,
    "soi_snd" / sockbuf_info,
    "soi_kind" / so_kind_t,
    "rfu_1" / uint32_t,  # reserved
    "soi_proto"
    / Switch(
        this.soi_kind,
        {
            so_kind_t.SOCKINFO_IN: Struct("pri_in" / in_sockinfo),
            so_kind_t.SOCKINFO_TCP: Struct("pri_tcp" / tcp_sockinfo),
            so_kind_t.SOCKINFO_UN: Struct("pri_un" / un_sockinfo),
            so_kind_t.SOCKINFO_NDRV: Struct("pri_ndrv" / ndrv_info),
            so_kind_t.SOCKINFO_KERN_EVENT: Struct("pri_kern_event" / kern_event_info),
            so_kind_t.SOCKINFO_KERN_CTL: Struct("pri_kern_ctl" / kern_ctl_info),
        },
    ),
)

socket_fdinfo = Struct(
    "pfi" / proc_fileinfo,
    "psi" / socket_info,
)

pipe_info = Struct(
    "pipe_stat" / vinfo_stat,
    "pipe_handle" / uint64_t,
    "pipe_peerhandle" / uint64_t,
    "pipe_status" / Int32sl,
    "rfu_1" / Int32sl,  # reserved
)

pipe_fdinfo = Struct(
    "pfi" / proc_fileinfo,
    "pipeinfo" / pipe_info,
)

vm_region_basic_info = Struct(
    "protection" / vm_prot_t,
    "max_protection" / vm_prot_t,
    "inheritance" / vm_inherit_t,
    "shared" / boolean_t,
    "reserved" / boolean_t,
    Padding(4),
    "offset" / memory_object_offset_t,
    "behavior" / vm_behavior_t,
    "user_wired_count" / Int16ul,
)

vm_region_basic_info_64 = Struct(
    "protection" / vm_prot_t,
    "max_protection" / vm_prot_t,
    "inheritance" / vm_inherit_t,
    "shared" / boolean_t,
    "reserved" / boolean_t,
    Padding(4),
    "offset" / memory_object_offset_t,
    "behavior" / vm_behavior_t,
    "user_wired_count" / Int16ul,
)

VM_REGION_BASIC_INFO_COUNT_64 = vm_region_basic_info_64.sizeof() // 4
vm_region_basic_info_data_t = vm_region_basic_info

task_dyld_info = Struct(
    "all_image_info_addr" / mach_vm_address_t,
    "all_image_info_size" / mach_vm_size_t,
    "all_image_info_format" / integer_t,
)

task_vm_info = Struct(
    "virtual_size" / mach_vm_size_t,
    "region_count" / integer_t,
    "page_size" / integer_t,
    "resident_size" / mach_vm_size_t,
    "resident_size_peak" / mach_vm_size_t,
    "device" / mach_vm_size_t,
    "device_peak" / mach_vm_size_t,
    "internal" / mach_vm_size_t,
    "internal_peak" / mach_vm_size_t,
    "external" / mach_vm_size_t,
    "external_peak" / mach_vm_size_t,
    "reusable" / mach_vm_size_t,
    "reusable_peak" / mach_vm_size_t,
    "purgeable_volatile_pmap" / mach_vm_size_t,
    "purgeable_volatile_resident" / mach_vm_size_t,
    "purgeable_volatile_virtual" / mach_vm_size_t,
    "compressed" / mach_vm_size_t,
    "compressed_peak" / mach_vm_size_t,
    "compressed_lifetime" / mach_vm_size_t,
    "phys_footprint" / mach_vm_size_t,
    "min_address" / mach_vm_address_t,
    "max_address" / mach_vm_address_t,
    "ledger_phys_footprint_peak" / Int64sl,
    "ledger_purgeable_nonvolatile" / Int64sl,
    "ledger_purgeable_novolatile_compressed" / Int64sl,
    "ledger_purgeable_volatile" / Int64sl,
    "ledger_purgeable_volatile_compressed" / Int64sl,
    "ledger_tag_network_nonvolatile" / Int64sl,
    "ledger_tag_network_nonvolatile_compressed" / Int64sl,
    "ledger_tag_network_volatile" / Int64sl,
    "ledger_tag_network_volatile_compressed" / Int64sl,
    "ledger_tag_media_footprint" / Int64sl,
    "ledger_tag_media_footprint_compressed" / Int64sl,
    "ledger_tag_media_nofootprint" / Int64sl,
    "ledger_tag_media_nofootprint_compressed" / Int64sl,
    "ledger_tag_graphics_footprint" / Int64sl,
    "ledger_tag_graphics_footprint_compressed" / Int64sl,
    "ledger_tag_graphics_nofootprint" / Int64sl,
    "ledger_tag_graphics_nofootprint_compressed" / Int64sl,
    "ledger_tag_neural_footprint" / Int64sl,
    "ledger_tag_neural_footprint_compressed" / Int64sl,
    "ledger_tag_neural_nofootprint" / Int64sl,
    "ledger_tag_neural_nofootprint_compressed" / Int64sl,
    "limit_bytes_remaining" / uint64_t,
    "decompressions" / integer_t,
    "ledger_swapins" / Int64sl,
    "ledger_tag_neural_nofootprint_total" / Int64sl,
    "ledger_tag_neural_nofootprint_peak" / Int64sl,
)

uuid_t = Struct(
    "time_low" / Int64ul,
    "time_mid" / Int32ul,
    "time_hi_and_version" / Int32ul,
    "clock_seq_hi_and_reserved" / Int8ul,
    "clock_seq_low" / Int8ul,
    "node" / Array(6, Int8ul),
)

dyld_image_info_t = Struct("imageLoadAddress" / Int64ul, "imageFilePath" / Int64ul, "imageFileModDate" / Int64ul)

all_image_infos_t = Struct(
    "version" / Int32ul,
    "infoArrayCount" / Int32ul,
    "infoArray" / Int64ul,
)

task_dyld_info_data_t = task_dyld_info
TASK_DYLD_INFO_COUNT = task_dyld_info_data_t.sizeof() / natural_t.sizeof()

task_vm_info_data_t = task_vm_info
TASK_VM_INFO_COUNT = task_vm_info_data_t.sizeof() / natural_t.sizeof()

STRUCT_ARM_THREAD_STATE64 = Struct(
    "x" / Array(29, uint64_t),
    "fp" / Default(uint64_t, 0),
    "lr" / Default(uint64_t, 0),
    "sp" / Default(uint64_t, 0),
    "pc" / Default(uint64_t, 0),
    "cpsr" / Default(uint32_t, 0),
    Padding(4),
)

arm_thread_state64_t = STRUCT_ARM_THREAD_STATE64
ARM_THREAD_STATE64_COUNT = arm_thread_state64_t.sizeof() // uint32_t.sizeof()

STRUCT_X86_THREAD_STATE32 = Struct(
    "eax" / Int32sl,
    "ebx" / Int32sl,
    "ecx" / Int32sl,
    "edx" / Int32sl,
    "edi" / Int32sl,
    "esi" / Int32sl,
    "ebp" / Int32sl,
    "esp" / Int32sl,
    "ss" / Int32sl,
    "eflags" / Int32sl,
    "eip" / Int32sl,
    "cs" / Int32sl,
    "ds" / Int32sl,
    "es" / Int32sl,
    "fs" / Int32sl,
    "gs" / Int32sl,
)

x86_thread_state32_t = STRUCT_X86_THREAD_STATE32

STRUCT_X86_THREAD_STATE64 = Struct(
    "rax" / uint64_t,
    "rbx" / uint64_t,
    "rcx" / uint64_t,
    "rdx" / uint64_t,
    "rdi" / uint64_t,
    "rsi" / uint64_t,
    "rbp" / uint64_t,
    "rsp" / uint64_t,
    "r8" / uint64_t,
    "r9" / uint64_t,
    "r10" / uint64_t,
    "r11" / uint64_t,
    "r12" / uint64_t,
    "r13" / uint64_t,
    "r14" / uint64_t,
    "r15" / uint64_t,
    "rip" / uint64_t,
    "rflags" / uint64_t,
    "cs" / uint64_t,
    "fs" / uint64_t,
    "gs" / uint64_t,
)

x86_thread_state64_t = STRUCT_X86_THREAD_STATE64

suseconds_t = uint32_t

timeval = Struct(
    "tv_sec" / time_t,
    Padding(4),
    "tv_usec" / suseconds_t,
    Padding(4),
)

procargs2_t = Struct(
    "argc" / Int32ul,
    "executable" / Aligned(4, CString("utf8")),
    "argv" / Array(this.argc, CString("utf8")),
    "_environ_apple" / GreedyRange(CString("utf8")),
    "environ_apple" / Computed(lambda ctx: [s for s in ctx._environ_apple if s]),
)

# See: https://opensource.apple.com/source/xnu/xnu-7195.81.3/EXTERNAL_HEADERS/mach-o/loader.h
LC_REQ_DYLD = 0x80000000

LOAD_COMMAND_TYPE = Enum(
    Int32ul,
    LC_SEGMENT=0x1,
    LC_SYMTAB=0x2,
    LC_SYMSEG=0x3,
    LC_THREAD=0x4,
    LC_UNIXTHREAD=0x5,
    LC_LOADFVMLIB=0x6,
    LC_IDFVMLIB=0x7,
    LC_IDENT=0x8,
    LC_FVMFILE=0x9,
    LC_PREPAGE=0xA,
    LC_DYSYMTAB=0xB,
    LC_LOAD_DYLIB=0xC,
    LC_ID_DYLIB=0xD,
    LC_LOAD_DYLINKER=0xE,
    LC_ID_DYLINKER=0xF,
    LC_PREBOUND_DYLIB=0x10,
    LC_ROUTINES=0x11,
    LC_SUB_FRAMEWORK=0x12,
    LC_SUB_UMBRELLA=0x13,
    LC_SUB_CLIENT=0x14,
    LC_SUB_LIBRARY=0x15,
    LC_TWOLEVEL_HINTS=0x16,
    LC_PREBIND_CKSUM=0x17,
    LC_LOAD_WEAK_DYLIB=(0x18 | LC_REQ_DYLD),
    LC_SEGMENT_64=0x19,
    LC_ROUTINES_64=0x1A,
    LC_UUID=0x1B,
    LC_RPATH=(0x1C | LC_REQ_DYLD),
    LC_CODE_SIGNATURE=0x1D,
    LC_SEGMENT_SPLIT_INFO=0x1E,
    LC_REEXPORT_DYLIB=0x1F | LC_REQ_DYLD,
    LC_LAZY_LOAD_DYLIB=0x20,
    LC_ENCRYPTION_INFO=0x21,
    LC_DYLD_INFO=0x22,
    LC_DYLD_INFO_ONLY=(0x22 | LC_REQ_DYLD),
    LC_LOAD_UPWARD_DYLIB=(0x23 | LC_REQ_DYLD),
    LC_VERSION_MIN_MACOSX=0x24,
    LC_VERSION_MIN_IPHONEOS=0x25,
    LC_FUNCTION_STARTS=0x26,
    LC_DYLD_ENVIRONMENT=0x27,
    LC_MAIN=(0x28 | LC_REQ_DYLD),
    LC_DATA_IN_CODE=0x29,
    LC_SOURCE_VERSION=0x2A,
    LC_DYLIB_CODE_SIGN_DRS=0x2B,
    LC_ENCRYPTION_INFO_64=0x2C,
    LC_LINKER_OPTION=0x2D,
    LC_LINKER_OPTIMIZATION_HINT=0x2E,
    LC_VERSION_MIN_TVOS=0x2F,
    LC_VERSION_MIN_WATCHOS=0x30,
    LC_NOTE=0x31,
    LC_BUILD_VERSION=0x32,
    LC_DYLD_EXPORTS_TRIE=(0x33 | LC_REQ_DYLD),
    LC_DYLD_CHAINED_FIXUPS=(0x34 | LC_REQ_DYLD),
    LC_FILESET_ENTRY=(0x35 | LC_REQ_DYLD),
)

# Load commands:
# reference - https://opensource.apple.com/source/xnu/xnu-2050.18.24/EXTERNAL_HEADERS/mach-o/loader.h

FAT_MAGIC = 0xCAFEBABE
FAT_CIGAM = 0xBEBAFECA

cpu_type_t = Int32ul
cpu_subtype_t = Int32ul

version_t = BitStruct("major" / BitsInteger(16), "minor" / Octet, "bug" / Octet)

segment_command_t = Struct(
    "segname" / PaddedString(16, "utf8"),
    "vmaddr" / Int64ul,
    "vmsize" / Int64ul,
    "fileoff" / Int64ul,
    "filesize" / Int64ul,
    "maxprot" / Int32ul,
    "initprot" / Int32ul,
    "nsects" / Int32ul,
    "flags" / Int32ul,
)

uuid_command_t = Struct("uuid" / Array(16, Int8ul))

build_tool_version = Struct("tool" / Int32ul, "version" / Int32ul)

build_version_command_t = Struct(
    "clients" / Int32ul,
    "minos" / version_t,
    "sdk" / version_t,
    "ntools" / Int32ul,
)

encryption_info_command = Struct(
    "cryptoff" / Int32ul,  # file offset of encrypted range
    "cryptsize" / Int32ul,  # file size of encrypted range
    "cryptid" / Int32ul,  # which encryption system, 0 means not-encrypted yet
)

encryption_info_command_64 = Struct(
    "cryptoff" / Int32ul,  # file offset of encrypted range
    "cryptsize" / Int32ul,  # file size of encrypted range
    "cryptid_offset" / Tell,
    "cryptid" / Int32ul,  # which encryption system, 0 means not-encrypted yet
    "_pad" / Int32ul,
)

load_command_t = Struct(
    "_start" / Tell,
    "cmd" / LOAD_COMMAND_TYPE,
    "cmdsize" / Int32ul,
    "_data_offset" / Tell,
    "data"
    / Switch(
        this.cmd,
        {
            LOAD_COMMAND_TYPE.LC_BUILD_VERSION: build_version_command_t,
            LOAD_COMMAND_TYPE.LC_UUID: uuid_command_t,
            LOAD_COMMAND_TYPE.LC_SEGMENT_64: segment_command_t,
            LOAD_COMMAND_TYPE.LC_ENCRYPTION_INFO_64: encryption_info_command_64,
            LOAD_COMMAND_TYPE.LC_ENCRYPTION_INFO: encryption_info_command,
        },
        Bytes(this.cmdsize - (this._data_offset - this._start)),
    ),
    Seek(this._start + this.cmdsize),
)

mach_header_t = Struct(
    "load_address" / Tell,
    "magic" / Hex(Int32ul),
    "cputype" / Hex(cpu_type_t),
    "cpusubtype" / Hex(cpu_subtype_t),
    "filetype" / Hex(Int32ul),
    "ncmds" / Hex(Int32ul),
    "sizeofcmds" / Hex(Int32ul),
    "flags" / Hex(Int32ul),
    "reserved" / Hex(Int32ul),
    "load_commands" / LazyArray(this.ncmds, load_command_t),
)

fat_arch = Struct(
    "cputype" / Hex(Int32ub),
    "cpusubtype" / Hex(Int32ub),
    "offset" / Hex(Int32ub),
    "size" / Hex(Int32ub),
    "align" / Hex(Int32ub),
)

fat_header = Struct("magic" / Hex(Int32ul), "nfat_arch" / Hex(Int32ub), "archs" / Array(this.nfat_arch, fat_arch))

thread_identifier_info = Struct(
    "thread_id" / Int64ul,  # system-wide unique 64-bit thread id
    "thread_handle" / Int64ul,  # handle to be used by libproc
    "dispatch_qaddr" / Int64ul,  # libdispatch queue address
)

THREAD_IDENTIFIER_INFO_COUNT = thread_identifier_info.sizeof() / natural_t.sizeof()


ipc_info_name_t = Struct(
    "iin_name" / mach_port_name_t,  # port name, including gen number
    "iin_collision" / integer_t,  # collision at this entry?
    "iin_type" / mach_port_type_t,  # straight port type
    "iin_urefs" / mach_port_urefs_t,  # user-references
    "iin_object" / natural_t,  # object pointer/identifier
    "iin_next" / natural_t,  # marequest/next in free list
    "iin_hash" / natural_t,  # hash index
)

POLLIN = 0x01
POLLPRI = 0x02
POLLOUT = 0x04
POLLERR = 0x08
POLLHUP = 0x10
POLLNVAL = 0x20


pollfd = Struct(
    "fd" / Int32sl,
    "events" / Int16sl,
    "revents" / Int16sl,
)



import tempfile
from pathlib import Path
from stat import S_IMODE

import pytest

from rpcclient.clients.darwin.consts import UF_IMMUTABLE
from rpcclient.core.structs.consts import LOCK_EX, LOCK_NB, LOCK_UN
from rpcclient.exceptions import RpcFileNotFoundError, RpcPermissionError


def test_touch(client, tmp_path):
    file = tmp_path / "temp.txt"
    client.fs.touch(file, mode=0o666)
    umask = client.symbols.umask(0o22)
    client.symbols.umask(umask)
    assert S_IMODE(client.fs.stat(file).st_mode) == 0o666 & ~umask


def test_chown(client, tmp_path):
    file = tmp_path / "temp.txt"
    client.fs.touch(file)
    pre_chown_stat = client.fs.stat(file)
    client.fs.chown(file, 2, -1)
    post_chown_stat = client.fs.stat(file)
    assert post_chown_stat.st_uid == 2
    assert pre_chown_stat.st_gid == post_chown_stat.st_gid


def test_chmod(client, tmp_path):
    file = tmp_path / "temp.txt"
    client.fs.touch(file, mode=0o777)
    client.fs.chmod(file, 0o666)
    assert S_IMODE(client.fs.stat(file).st_mode) == 0o666


def test_open(client, tmp_path):
    file = tmp_path / "temp.txt"
    with client.fs.open(file, "rw", 0o666):
        pass
    umask = client.symbols.umask(0o22)
    client.symbols.umask(umask)
    assert S_IMODE(client.fs.stat(file).st_mode) == 0o666 & ~umask


def test_flock(client, tmp_path):
    file = tmp_path / "lock.txt"
    with client.fs.open(file, "w+") as f:
        f.flock(LOCK_EX | LOCK_NB)
        f.flock(LOCK_UN)


def test_remove(client, tmp_path):
    file = tmp_path / "temp.txt"
    client.fs.touch(file)
    client.fs.remove(file)
    assert not client.fs.accessible(file)


def test_mkdir(client, tmp_path):
    dir_ = tmp_path / "test_dir"
    client.fs.mkdir(dir_, 0o666)
    assert client.fs.accessible(dir_)


def test_chdir_pwd(client, tmp_path):
    client.fs.chdir(tmp_path)
    assert client.fs.pwd() == str(tmp_path)


def test_symlink(client, tmp_path):
    file = tmp_path / "temp.txt"
    symlink = tmp_path / "temp1.txt"
    client.fs.touch(file)
    client.fs.symlink(file, symlink)
    assert client.fs.accessible(symlink)
    assert client.fs.readlink(symlink) == str(file)


def test_realpath(client, tmp_path):
    assert client.fs.realpath(tmp_path / "././") == str(tmp_path)


def test_link(client, tmp_path):
    client.fs.write_file(tmp_path / "temp.txt", b"hello")
    client.fs.link((tmp_path / "temp.txt"), (tmp_path / "temp1.txt"))
    assert client.fs.read_file(tmp_path / "temp1.txt") == b"hello"


def test_listdir(client, tmp_path):
    assert not client.fs.listdir(tmp_path)
    client.fs.touch(tmp_path / "temp.txt")
    assert client.fs.listdir(tmp_path) == ["temp.txt"]


def test_listdir_non_exists(client):
    with pytest.raises(RpcFileNotFoundError):
        client.fs.listdir("/non_exists_path")


def test_push_expand_path(client, tmp_path):
    with tempfile.TemporaryDirectory() as temp_dir:
        local_file = Path(temp_dir) / "local"
        local_file.touch()
        remote_file = tmp_path / "temp.txt"
        client.fs.push(local_file, remote_file)
        assert remote_file.exists()


def test_pull_expand_path(client, tmp_path):
    with tempfile.TemporaryDirectory() as temp_dir:
        remote_file = tmp_path / "temp.txt"
        remote_file.touch()
        local_file = Path(temp_dir) / "local"
        client.fs.pull(remote_file, local_file)
        assert local_file.exists()


@pytest.mark.parametrize("file_size", [1024, 10240, 102400])
def test_push_pull_with_different_sizes(client, tmp_path, file_size):
    assert not client.fs.listdir(tmp_path)
    local = Path("/tmp/temp.bin")
    local_pull = Path("/tmp/temp2.bin")
    remote = tmp_path / "temp.bin"

    with open(local, "wb") as f:
        f.write(b"\0" * file_size)

    client.fs.push(local, remote)

    assert client.fs.lstat(remote).st_size == file_size
    client.fs.pull(remote, local_pull)
    assert local_pull.stat().st_size == file_size
    local.unlink(missing_ok=True)
    local_pull.unlink(missing_ok=True)


def test_pull(client, tmp_path):
    client.fs.touch(tmp_path / "a")
    with tempfile.TemporaryDirectory() as local_dir:
        local_dir = Path(local_dir)
        client.fs.pull(tmp_path / "a", local_dir)
        assert (local_dir / "a").exists()
    with tempfile.TemporaryDirectory() as local_dir:
        local_dir = Path(local_dir)
        client.fs.pull(tmp_path / "a", local_dir / "a")
        assert (local_dir / "a").exists()

    client.fs.mkdir(tmp_path / "b")
    with tempfile.TemporaryDirectory() as local_dir:
        local_dir = Path(local_dir)
        client.fs.pull(tmp_path / "b", local_dir, recursive=True)
        assert (local_dir / "b").exists()
        client.fs.pull(tmp_path / "b", local_dir / "b", recursive=True)
        assert (local_dir / "b" / "b").exists()


def test_push_pull_dir(client, tmp_path: Path):
    (tmp_path / "a").touch()
    with tempfile.TemporaryDirectory() as local_dir:
        local_dir = Path(local_dir)
        client.fs.pull(tmp_path, local_dir, recursive=True)
        assert (local_dir / tmp_path.name / "a").exists()
    with tempfile.TemporaryDirectory() as local_dir:
        local_dir = Path(local_dir)
        (local_dir / "b").touch()
        client.fs.push(local_dir, tmp_path, recursive=True)
        a = tmp_path / local_dir.name / "b"
        assert a.exists()


def test_scandir_sanity(client, tmp_path):
    entries = list(client.fs.scandir(tmp_path))
    assert not entries
    client.fs.write_file(tmp_path / "temp.txt", b"hello")
    entries = list(client.fs.scandir(tmp_path))
    assert len(entries) == 1
    assert entries[0].name == "temp.txt"
    assert entries[0].path == str(tmp_path / "temp.txt")
    assert entries[0].is_file()
    assert not entries[0].is_dir()
    assert not entries[0].is_symlink()


def test_stat_sanity(client, tmp_path):
    file = tmp_path / "temp.txt"
    client.fs.write_file(file, "h" * 0x10000)
    client_stat = client.fs.stat(file)
    path_stat = client.fs.stat(file)
    assert path_stat.st_dev == client_stat.st_dev
    assert path_stat.st_ino == client_stat.st_ino
    assert path_stat.st_mode == client_stat.st_mode
    assert path_stat.st_nlink == client_stat.st_nlink
    assert path_stat.st_uid == client_stat.st_uid
    assert path_stat.st_gid == client_stat.st_gid
    assert path_stat.st_rdev == client_stat.st_rdev
    assert path_stat.st_atime == client_stat.st_atime
    assert path_stat.st_mtime == client_stat.st_mtime
    assert path_stat.st_ctime == client_stat.st_ctime
    assert path_stat.st_size == client_stat.st_size
    assert path_stat.st_blocks == client_stat.st_blocks
    assert path_stat.st_blksize == client_stat.st_blksize
    assert path_stat.st_flags == client_stat.st_flags
    assert path_stat.st_gen == client_stat.st_gen


def test_walk(client, tmp_path):
    client.fs.mkdir(tmp_path / "dir_a")
    client.fs.touch(tmp_path / "dir_a" / "a1.txt")
    client.fs.touch(tmp_path / "dir_a" / "a2.txt")
    client.fs.mkdir(tmp_path / "dir_b")
    client.fs.touch(tmp_path / "dir_b" / "b1.txt")
    client.fs.touch(tmp_path / "dir_b" / "b2.txt")

    assert list(client.fs.walk(tmp_path)) == [
        (f"{tmp_path}", ["dir_b", "dir_a"], []),
        (f"{tmp_path}/dir_b", [], ["b2.txt", "b1.txt"]),
        (f"{tmp_path}/dir_a", [], ["a1.txt", "a2.txt"]),
    ]


@pytest.mark.darwin
def test_xattr(client, tmp_path):
    client.fs.setxattr(tmp_path, "KEY", b"VALUE")
    assert client.fs.getxattr(tmp_path, "KEY") == b"VALUE"
    assert client.fs.listxattr(tmp_path) == ["KEY"]
    assert client.fs.dictxattr(tmp_path) == {"KEY": b"VALUE"}
    client.fs.removexattr(tmp_path, "KEY")
    assert client.fs.listxattr(tmp_path) == []


@pytest.mark.darwin
def test_chflags(client, tmp_path):
    # create temporary file
    file = tmp_path / "file"
    with client.fs.open(file, "w"):
        pass
    # make it immutable
    client.fs.chflags(file, UF_IMMUTABLE)
    # verify we cannot remove it
    with pytest.raises(RpcPermissionError):
        client.fs.remove(file)
    # restore its permissions
    client.fs.chflags(file, 0)
    # verify removal succeeds
    client.fs.remove(file)



import logging
import struct
import threading
from socket import socket

from rpcclient.exceptions import ServerDiedError
from rpcclient.protos.rpc_pb2 import Handshake, ProtocolConstants, RpcMessage, RpcPtyMessage

logger = logging.getLogger(__name__)


class RpcSocket:
    """
    Facilitates communication with a remote server using sockets and implements
    specific RPC (Remote Procedure Call) messaging protocols.

    This class provides methods to send and receive RPC messages, ensuring
    correct serialization and deserialization of messages. Additionally, it
    manages a protocol lock for synchronized message send and receive operations
    where applicable.

    Attributes:
        raw_socket: The underlying socket used for communication with the remote server.
    """

    def __init__(self, sock: socket):
        self.raw_socket = sock
        self._protocol_lock = threading.Lock()

    def _msg_recv(self) -> bytes:
        try:
            size = struct.unpack("<Q", self.raw_socket.recv(8))[0]
            buff = self._recvall(size)
        except struct.error as e:
            raise ConnectionError() from e
        return buff

    def _msg_send(self, message: bytes) -> None:
        buff = struct.pack("<Q", len(message)) + message
        self.raw_socket.sendall(buff)

    def _recvall(self, size: int) -> bytes:
        buf = b""
        while size:
            try:
                chunk = self.raw_socket.recv(size)
            except BlockingIOError:
                continue
            if not self.raw_socket.getblocking() and not chunk:
                raise ServerDiedError()
            size -= len(chunk)
            buf += chunk
        return buf

    def rpc_handshake_recv(self):
        rpc_handshake = Handshake()
        rpc_handshake.ParseFromString(self._msg_recv())
        return rpc_handshake

    def rpc_msg_recv(self):
        rpc_msg = RpcMessage()
        rpc_msg.ParseFromString(self._msg_recv())
        return rpc_msg

    def rpc_msg_recv_pty(self):
        rpc_msg = RpcPtyMessage()
        rpc_msg.ParseFromString(self._msg_recv())
        return rpc_msg

    def rpc_msg_send(self, msg: RpcMessage) -> None:
        msg.magic = ProtocolConstants.MESSAGE_MAGIC
        rpc_msg = msg.SerializeToString()
        self._msg_send(rpc_msg)

    def rpc_msg_send_recv(self, msg: RpcMessage) -> RpcMessage:
        with self._protocol_lock:
            self.rpc_msg_send(msg)
            return self.rpc_msg_recv()



from rpcclient.clients.darwin.client import DarwinClient
from rpcclient.clients.darwin.subsystems.reports import Reports
from rpcclient.clients.ios.subsystems.accessibility import Accessibility
from rpcclient.clients.ios.subsystems.amfi import Amfi
from rpcclient.clients.ios.subsystems.backlight import Backlight
from rpcclient.clients.ios.subsystems.lockdown import Lockdown
from rpcclient.clients.ios.subsystems.mobile_gestalt import MobileGestalt
from rpcclient.clients.ios.subsystems.processes import IosProcesses
from rpcclient.clients.ios.subsystems.screen_capture import ScreenCapture
from rpcclient.clients.ios.subsystems.sprinboard import SpringBoard
from rpcclient.clients.ios.subsystems.telephony import Telephony
from rpcclient.clients.ios.subsystems.wifi import IosWifi
from rpcclient.core.subsystems.decorator import subsystem

CRASH_REPORTS_DIR = "Library/Logs/CrashReporter"


class IosClient(DarwinClient):
    @subsystem
    def backlight(self) -> Backlight:
        return Backlight(self)

    @subsystem
    def reports(self) -> Reports:
        return Reports(self, CRASH_REPORTS_DIR)

    @subsystem
    def mobile_gestalt(self) -> MobileGestalt:
        return MobileGestalt(self)

    @subsystem
    def processes(self) -> IosProcesses:
        return IosProcesses(self)

    @subsystem
    def lockdown(self) -> Lockdown:
        return Lockdown(self)

    @subsystem
    def telephony(self) -> Telephony:
        return Telephony(self)

    @subsystem
    def screen_capture(self) -> ScreenCapture:
        return ScreenCapture(self)

    @subsystem
    def accessibility(self) -> Accessibility:
        return Accessibility(self)

    @subsystem
    def wifi(self) -> IosWifi:
        return IosWifi(self)

    @subsystem
    def springboard(self) -> SpringBoard:
        return SpringBoard(self)

    @subsystem
    def amfi(self) -> Amfi:
        return Amfi(self)

    @property
    def roots(self) -> list[str]:
        """get a list of all accessible darwin roots when used for lookup of files/preferences/..."""
        return [*super().roots, "/var/mobile"]

    @property
    def airplane_mode(self) -> bool:
        # use MobileGestalt for a more accurate result
        return self.mobile_gestalt.AirplaneMode

    @airplane_mode.setter
    def airplane_mode(self, value: bool):
        """set whether the device should enter airplane mode (turns off baseband, bt, etc...)"""
        radio_preferences = self.symbols.objc_getClass("RadiosPreferences").objc_call("new")
        radio_preferences.objc_call("setAirplaneMode:", value)
        radio_preferences.objc_call("synchronize")



import logging
from typing import Any, Callable, Union

from rpcclient.exceptions import ServerDiedError, SubsystemInitError

logger = logging.getLogger(__name__)


class SubsystemNotAvailable:
    """Null-object for an unavailable subsystem."""

    __slots__ = ("name", "reason")

    def __init__(self, name: str, reason: Union[str, BaseException] = "Not supported") -> None:
        """Store the subsystem name and the failure reason."""
        self.name = name
        self.reason = str(reason)

    def __bool__(self) -> bool:
        """Always False so `if obj.subsystem:` is skipped."""
        return False

    def __getattr__(self, attr: str) -> Any:
        """Raise a clear error on any attribute access."""
        raise SubsystemInitError(f"Subsystem '{self.name}' is not available: {self.reason}")

    def __repr__(self) -> str:
        """Debug-friendly representation."""
        return f"<SubsystemNotAvailable {self.name}: {self.reason}>"


class subsystem:
    """Descriptor that lazily initializes a subsystem and caches the result."""

    def __init__(self, fget: Callable[[Any], Any]) -> None:
        """Capture the factory callable and set cache metadata."""
        self.fget = fget
        self.__doc__ = getattr(fget, "__doc__", None)
        self.name = fget.__name__
        self.cache_key = f"_{self.name}"

    def __set_name__(self, owner: type, name: str) -> None:
        """Bind the descriptor name and derive the cache key."""
        self.name = name
        self.cache_key = f"_{name}"

    def __get__(self, obj: Any, objtype: Union[type, None] = None) -> Any:
        """Build on first access, cache, and return; wrap failures as NotAvailable."""
        if obj is None:
            return self
        d = obj.__dict__
        if self.cache_key in d:
            return d[self.cache_key]
        try:
            value = self.fget(obj)
        except Exception as e:
            if isinstance(e, ServerDiedError):
                raise
            logger.exception("Subsystem %s failed to initialize", self.name)
            value = SubsystemNotAvailable(self.name, e)
        d[self.cache_key] = value
        return value



import logging
from typing import TYPE_CHECKING, Optional

from rpcclient.exceptions import RpcAppleScriptError

if TYPE_CHECKING:
    from rpcclient.clients.macos.client import MacosClient

logger = logging.getLogger(__name__)


class AppleScript:
    def __init__(self, client: "MacosClient"):
        """
        :param client: rpcclient.macos.MacosClient
        """
        self._client = client

    def execute(self, script: str) -> None:
        with self._client.safe_malloc(8) as error:
            error[0] = 0
            apple_script = (
                self._client.symbols.objc_getClass("NSAppleScript")
                .objc_call("alloc")
                .objc_call("initWithSource:", self._client.cf(script))
            )
            apple_script.objc_call("executeAndReturnError:", error)
            if error[0]:
                raise RpcAppleScriptError(error[0].objc_call("description").py())
            apple_script.objc_call("release")

    def say(self, message: str, voice: Optional[str] = None):
        script = f'say "{message}"'
        if voice is not None:
            script += f' using "{voice}"'
        self.execute(script)

    def beep(self, count: int) -> None:
        self.execute(f"beep {count}")



import ctypes
import logging
from typing import TYPE_CHECKING, Optional

from rpcclient.clients.darwin.common import CfSerializable
from rpcclient.clients.darwin.symbol import DarwinSymbol
from rpcclient.core.allocated import Allocated
from rpcclient.exceptions import BadReturnValueError, RpcClientException

if TYPE_CHECKING:
    from rpcclient.clients.ios.client import IosClient

logger = logging.getLogger(__name__)


class WifiSavedNetwork:
    def __init__(self, client: "IosClient", wifi_manager: DarwinSymbol, network: DarwinSymbol) -> None:
        self._client = client
        self._wifi_manager = wifi_manager
        self._network = network

    @property
    def ssid(self) -> bytes:
        """Return the network SSID as raw bytes."""
        return self.get_property("SSID")

    @property
    def bssid(self) -> str:
        """Return the network BSSID as a string."""
        return self.get_property("BSSID")

    def get_property(self, name: str) -> CfSerializable:
        """Return a WiFiNetwork property by name from the underlying CF object."""
        return self._client.symbols.WiFiNetworkGetProperty(self._network, self._client.cf(name)).py()

    def forget(self) -> None:
        """Remove this network from the saved networks list."""
        self._client.symbols.WiFiManagerClientRemoveNetwork(self._wifi_manager, self._network)

    def __repr__(self) -> str:
        result = f"<{self.__class__.__name__} "
        result += f"SSID:{self.ssid} "
        result += f"BSSID:{self.bssid}"
        result += ">"
        return result


class WifiScannedNetwork:
    def __init__(self, client: "IosClient", interface: DarwinSymbol, network: dict) -> None:
        self._client = client
        self._interface = interface
        self.network = network

    @property
    def ssid(self) -> bytes:
        """Return the scanned network SSID as raw bytes."""
        return self.network.get("SSID")

    @property
    def bssid(self) -> str:
        """Return the scanned network BSSID as a string."""
        return self.network.get("BSSID")

    @property
    def rssi(self) -> int:
        """Return the received signal strength indicator (RSSI)."""
        return ctypes.c_int64(self.network["RSSI"]).value

    @property
    def channel(self) -> int:
        """Return the Wi-Fi channel number."""
        return self.network["CHANNEL"]

    def connect(self, password: Optional[str] = None) -> None:
        """Associate to this network, optionally providing a password."""
        result = self._client.symbols.Apple80211Associate(
            self._interface, self._client.cf(self.network), self._client.cf(password) if password else 0
        )

        if result:
            raise BadReturnValueError(f"Apple80211Associate() failed with: {result}")

    def disconnect(self) -> None:
        """Disassociate from the current network."""
        if self._client.symbols.Apple80211Disassociate(self._interface):
            raise BadReturnValueError("Apple80211Disassociate() failed")

    def __repr__(self) -> str:
        result = f"<{self.__class__.__name__} "
        result += f"SSID:{self.ssid} "
        result += f"BSSID:{self.bssid} "
        result += f"CHANNEL:{self.channel} "
        result += f"RSSI:{self.rssi}"
        result += ">"
        return result


class WifiInterface(Allocated):
    def __init__(self, client: "IosClient", interface: DarwinSymbol) -> None:
        super().__init__()
        self._client = client
        self._interface = interface

    def scan(self, options: Optional[dict] = None) -> list[WifiScannedNetwork]:
        """Scan for nearby networks and return a list of WifiScannedNetwork."""

        if options is None:
            options = {}

        result = []
        with self._client.safe_malloc(8) as p_found_networks:
            while True:
                scan_result = self._client.symbols.Apple80211Scan(
                    self._interface, p_found_networks, self._client.cf(options)
                )
                if scan_result == 0:
                    break
                elif scan_result == -1:
                    raise BadReturnValueError("Apple80211Scan failed")
                # else, try again

            for network in p_found_networks[0].py():
                result.append(WifiScannedNetwork(self._client, self._interface, network))

        return result

    def disconnect(self) -> None:
        """Disconnect from the currently associated Wi-Fi network."""
        self._client.symbols.Apple80211Disassociate(self._interface)

    def _deallocate(self) -> None:
        self._client.symbols.Apple80211Close(self._interface)


class IosWifi:
    """iOS Wi-Fi utilities backed by WiFiKit and Apple80211."""

    def __init__(self, client: "IosClient") -> None:
        self._client = client
        self._client.load_framework("WiFiKit")

        self._wifi_manager = self._client.symbols.WiFiManagerClientCreate(0, 0)
        if not self._wifi_manager:
            self._client.raise_errno_exception("WiFiManagerClientCreate failed")

    @property
    def saved_networks(self) -> list[WifiSavedNetwork]:
        """Return saved Wi-Fi networks known to the device."""
        result = []

        network_list = self._client.symbols.WiFiManagerClientCopyNetworks(self._wifi_manager).py()
        if not network_list:
            return result

        for network in network_list:
            result.append(WifiSavedNetwork(self._client, self._wifi_manager, network))

        return result

    @property
    def interfaces(self) -> list[str]:
        """Return a list of available Wi-Fi interface names."""
        with self._client.safe_malloc(8) as p_interface:
            if self._client.symbols.Apple80211Open(p_interface):
                self._client.raise_errno_exception("Apple80211Open() failed")

            with self._client.safe_malloc(8) as p_interface_names:
                if self._client.symbols.Apple80211GetIfListCopy(p_interface[0], p_interface_names):
                    self._client.raise_errno_exception("Apple80211GetIfListCopy() failed")

                return p_interface_names[0].py()

    def turn_on(self) -> None:
        """Enable Wi-Fi on the device."""
        self._set(True)

    def turn_off(self) -> None:
        """Disable Wi-Fi on the device."""
        self._set(False)

    def is_on(self) -> bool:
        """Return True if Wi-Fi is enabled."""
        return bool(
            self._client.symbols.WiFiManagerClientCopyProperty(self._wifi_manager, self._client.cf("AllowEnable")).py()
        )

    def get_interface(self, interface_name: Optional[str] = None) -> WifiInterface:
        """Return a bound interface handle for scanning and associating."""
        with self._client.safe_malloc(8) as p_handle:
            if self._client.symbols.Apple80211Open(p_handle):
                self._client.raise_errno_exception("Apple80211Open() failed")
            handle = p_handle[0]

        if interface_name is None:
            wifi_interfaces = self.interfaces

            if not wifi_interfaces:
                raise RpcClientException("no available wifi interfaces were found")

            interface_name = wifi_interfaces[0]

        if self._client.symbols.Apple80211BindToInterface(handle, self._client.cf(interface_name)):
            self._client.raise_errno_exception("Apple80211BindToInterface failed")

        return WifiInterface(self._client, handle)

    def _set(self, is_on: bool) -> None:
        if not self._client.symbols.WiFiManagerClientSetProperty(
            self._wifi_manager, self._client.cf("AllowEnable"), self._client.cf(is_on)
        ):
            raise BadReturnValueError("WiFiManagerClientSetProperty() failed")



import contextlib

import pytest

from rpcclient.exceptions import RpcPermissionError

pytestmark = pytest.mark.darwin


def test_query_certificates(client):
    with contextlib.suppress(RpcPermissionError):
        assert len(client.keychain.query_certificates()) > 1


def test_query_keys(client):
    with contextlib.suppress(RpcPermissionError):
        assert len(client.keychain.query_keys()) > 1



import dataclasses
import time
from collections.abc import Iterator
from enum import IntEnum, IntFlag
from typing import TYPE_CHECKING, Any, Optional

from rpcclient.clients.darwin.symbol import DarwinSymbol
from rpcclient.exceptions import (
    ElementNotFoundError,
    FirstElementNotFoundError,
    LastElementNotFoundError,
    RpcAccessibilityTurnedOffError,
    RpcFailedToGetPrimaryAppError,
)

if TYPE_CHECKING:
    from rpcclient.clients.ios.client import IosClient


class AXDirection(IntEnum):
    Next = 1
    Previous = 2


class FrameStyle(IntEnum):
    Default = 0
    Blue = 0
    Green = 1
    Yellow = 4
    LightGreen = 5
    Invisible = 6


class AXTraits(IntFlag):
    kAXButtonTrait = 0x01
    kAXLinkTrait = 0x02
    kAXVisitedTrait = 0x100000000
    kAXHeaderTrait = 0x10000
    kAXFooterTrait = 0x4000000
    kAXWebContentTrait = 0x20000
    kAXTextEntryTrait = 0x40000
    kAXImageTrait = 0x4
    kAXSelectedTrait = 0x8
    kAXPlaysSoundTrait = 0x10
    kAXKeyboardKeyTrait = 0x20
    kAXStaticTextTrait = 0x40
    kAXSummaryElementTrait = 0x80
    kAXNotEnabledTrait = 0x100
    kAXPickerElementTrait = 0x80000
    kAXUpdatesFrequentlyTrait = 0x200
    kAXIsEditingTrait = 0x200000
    kAXLaunchIconTrait = 0x400000
    kAXFolderIconTrait = 0x4000000000000
    kAXSearchFieldTrait = 0x400
    kAXStatusBarElementTrait = 0x800000
    kAXAllowsLayoutChangeInStatusBarTrait = 0x400000000000000
    kAXSecureTextFieldTrait = 0x1000000
    kAXBackButtonTrait = 0x8000000
    kAXToggleTrait = 0x20000000000000
    kAXSelectionDismissesItemTrait = 0x80000000
    kAXScrollableTrait = 0x200000000
    kAXSpacerTrait = 0x400000000
    kAXTableIndexTrait = 0x800000000
    kAXMapTrait = 0x1000000000
    kAXTextOperationsAvailableTrait = 0x2000000000
    kAXDraggableTrait = 0x4000000000
    kAXGesturePracticeRegionTrait = 0x8000000000
    kAXPopupButtonTrait = 0x10000000000
    kAXAllowsNativeSlidingTrait = 0x20000000000
    kAXTouchContainerTrait = 0x200000000000
    kAXSupportsZoomTrait = 0x400000000000
    kAXTextAreaTrait = 0x800000000000
    kAXBookContentTrait = 0x1000000000000
    kAXStartsMediaSessionTrait = 0x800
    kAXAdjustableTrait = 0x1000
    kAXMenuItemTrait = 0x10000000000000
    kAXAutoCorrectCandidateTrait = 0x20000000
    kAXDeleteKeyTrait = 0x40000000
    kAXTabButtonTrait = 0x10000000
    kAXIgnoreItemChooserTrait = 0x40000000000000
    kAXAllowsDirectInteractionTrait = 0x2000
    kAXCausesPageTurnTrait = 0x4000
    kAXTabBarTrait = 0x8000
    kAXRadioButtonTrait = 0x100000
    kAXMathEquationTrait = 0x40000000000
    kAXInactiveTrait = 0x2000000
    kAXSupportsTrackingDetailTrait = 0x80000000000000
    kAXAlertTrait = 0x100000000000000
    kAXReadOnlyTrait = 0x8000000000000
    kAXProminentIconTrait = 0x1000000000000000


@dataclasses.dataclass
class CGPoint:
    x: float
    y: float

    def __str__(self) -> str:
        return f"{{{self.x}, {self.y}}}"


@dataclasses.dataclass
class CGSize:
    width: float
    height: float

    def __str__(self) -> str:
        return f"{{{self.width}, {self.height}}}"


@dataclasses.dataclass
class CGRect:
    origin: CGPoint
    size: CGSize

    def __str__(self) -> str:
        return f"{{{self.origin}, {self.size}}}"


class AXElement(DarwinSymbol):
    """
    Wrapper around the device AXElement Objective-C object.

    Built from reversed XADInspectorManager methods.
    """

    @property
    def first_element(self) -> "AXElement":
        """Return the first element in the accessibility hierarchy."""
        result = self._element_for_attribute(3000)
        if not result:
            raise FirstElementNotFoundError("failed to get first element in hierarchy")

        if result.ui_element.objc_call("boolWithAXAttribute:", 2046):
            result = result._element_for_attribute(3000)
        return result

    @property
    def last_element(self) -> "AXElement":
        """Return the last element in the accessibility hierarchy."""
        result = self._element_for_attribute(3016)
        if not result:
            raise LastElementNotFoundError("failed to get last element in hierarchy")

        if result.ui_element.objc_call("boolWithAXAttribute:", 2046):
            result = result._element_for_attribute(3016)
        return result

    @property
    def identifier(self) -> str:
        """Return the element identifier."""
        return self.objc_call("identifier").py(encoding="utf8")

    @property
    def url(self) -> str:
        """Return the element URL."""
        return self.objc_call("url").py(encoding="utf8")

    @property
    def path(self) -> DarwinSymbol:
        """Return the element path object."""
        return self.objc_call("path")

    @property
    def frame(self) -> CGRect:
        """Return the element frame as a CGRect."""
        result = self.objc_call("frame", return_raw=True)
        return CGRect(origin=CGPoint(x=result.d0, y=result.d1), size=CGSize(width=result.d2, height=result.d3))

    @property
    def label(self) -> Optional[str]:
        """Return the visible label text."""
        return self.objc_call("label").py(encoding="utf8")

    @property
    def value(self) -> str:
        """Return the element value."""
        return self.objc_call("value").py(encoding="utf8")

    @property
    def bundle_identifier(self) -> str:
        """Return the owning app bundle identifier."""
        return self.objc_call("bundleId").py(encoding="utf8")

    @property
    def pid(self) -> int:
        """Return the owning process ID."""
        return self.objc_call("pid").c_uint16

    @property
    def process_name(self) -> str:
        """Return the owning process name."""
        return self.objc_call("processName").py()

    @property
    def screen_locked(self) -> bool:
        """Return True if the screen is locked."""
        return self.objc_call("isScreenLocked") == 1

    @property
    def is_accessibility_opaque_element_provider(self) -> bool:
        """Return True if the element provides its own accessibility hierarchy."""
        return self.objc_call("isAccessibilityOpaqueElementProvider") != 0

    @property
    def parent(self) -> Optional["AXElement"]:
        """Return the parent element if available."""
        tmp = self._element_for_attribute(2066)
        if tmp:
            return tmp

        tmp = self._element_for_attribute(2092)
        if tmp:
            return tmp.parent

        return None

    @property
    def ui_element(self) -> DarwinSymbol:
        """Return the underlying AXUIElement."""
        return self.objc_call("uiElement")

    @property
    def traits(self) -> AXTraits:
        """Return the element traits as AXTraits."""
        return AXTraits(self.objc_call("traits").c_uint64)

    @property
    def elements(self) -> list["AXElement"]:
        """Return the list of currently displayed elements."""
        result = []
        elements = self.objc_call("explorerElements")
        for i in range(elements.objc_call("count")):
            result.append(self._client.accessibility.axelement(elements.objc_call("objectAtIndex:", i)))
        return result

    def insert_text(self, text: str) -> None:
        """Insert text into the focused editable element."""
        self.objc_call("insertText:", self._client.cf(text))

    def delete_text(self) -> None:
        """Delete a character from the focused editable element."""
        self.objc_call("deleteText")

    def highlight(self) -> None:
        """Draw a frame around the element, replacing any existing frame."""
        frame = self.frame
        self._client.accessibility.draw_frame(frame.origin.x, frame.origin.y, frame.size.width, frame.size.height)

    def scroll_to_visible(self) -> None:
        """Scroll until the element becomes fully visible."""
        self.objc_call("scrollToVisible")

    def press(self) -> None:
        """Activate/press the element."""
        self.objc_call("press")

    def long_press(self) -> None:
        """Long-press the element."""
        self.objc_call("longPress")

    def __iter__(self) -> Iterator["AXElement"]:
        current = self.first_element
        while current:
            yield current
            current = current.next()

    def _element_for_attribute(self, axattribute: int, parameter: Optional[Any] = None) -> Optional["AXElement"]:
        if parameter is None:
            result = self.objc_call("elementForAttribute:", axattribute)
        else:
            result = self.objc_call("elementForAttribute:parameter:", axattribute, parameter)
        return AXElement.create(result, self._client)

    def _next_opaque(self, direction: AXDirection = AXDirection.Next) -> Optional["AXElement"]:
        element = self

        if not element.is_accessibility_opaque_element_provider:
            element = self.parent

        if not element:
            return None

        element = element._element_for_attribute(
            95225,
            self._client.cf([
                direction,
                0,
                self._client.symbols.objc_getClass("NSValue").objc_call("valueWithRange:", 0x7FFFFFFF, 0),
                "AXAudit",
            ]),
        )

        if element:
            ui_element = element.ui_element
            if ui_element and ui_element.objc_call("boolWithAXAttribute:", 2046):
                return element._next_opaque

        return element

    def _next_elements_with_count(self, count: int) -> list["AXElement"]:
        elements = self.objc_call("nextElementsWithCount:", count)
        result = []
        for i in range(elements.objc_call("count")):
            result.append(AXElement.create(elements.objc_call("objectAtIndex:", i), self._client))
        return result

    def _previous_elements_with_count(self, count: int) -> list["AXElement"]:
        elements = self.objc_call("previousElementsWithCount:", count)
        result = []
        for i in range(elements.objc_call("count")):
            result.append(AXElement.create(elements.objc_call("objectAtIndex:", i), self._client))
        return result

    def _set_assistive_focus(self, focused: bool) -> None:
        self.ui_element.objc_call(
            "setAXAttribute:withObject:synchronous:",
            2018,
            self._client.cf({"focused": int(focused), "assistiveTech": "AXAudit"}),
            0,
        )
        parent = self._element_for_attribute(2092)
        if parent:
            parent._set_assistive_focus(focused)

    def next(self, direction: AXDirection = AXDirection.Next, cyclic: bool = False) -> Optional["AXElement"]:
        """
        Return and scroll to the next element in the current view.

        This method was created by reversing [XADInspectorManager _nextElementNavigationInDirection:forElement:]
        so we don't really know much about the used consts.
        """
        next_opaque = self._next_opaque(direction)

        if not self.is_accessibility_opaque_element_provider and next_opaque:
            return next_opaque

        if direction == AXDirection.Next:
            next_or_prev_list = self._next_elements_with_count(1)
        else:
            next_or_prev_list = self._previous_elements_with_count(1)

        if next_or_prev_list:
            result = next_or_prev_list[0]
            if result.is_accessibility_opaque_element_provider:
                focused_element = self._element_for_attribute(95226, self._client.cf("AXAudit"))
                if focused_element:
                    focused_element._set_assistive_focus(False)
                result._set_assistive_focus(False)
                result = result._next_opaque(direction)

            if result and not result.is_accessibility_opaque_element_provider:
                return result

        result = self._next_opaque(direction)
        if result:
            return result

        if not self.is_accessibility_opaque_element_provider:
            parent = self.parent
            if parent:
                return parent.next(direction, cyclic=cyclic)

        if cyclic:
            if direction == AXDirection.Next:
                return self._client.accessibility.primary_app.first_element
            return self._client.accessibility.primary_app.last_element

        return None

    def compare_label(self, label: str, auto_scroll: bool = True, draw_frame: bool = True) -> bool:
        """
        Compare a label against this element's label.

        Optionally, scroll to the element and draw a highlight frame.
        """
        if auto_scroll:
            self.scroll_to_visible()

        if draw_frame:
            self.highlight()

        return self.label == label

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} LABEL:{self.label}>"

    def __str__(self) -> str:
        result = self.label
        return result if result else "NO LABEL"


class Accessibility:
    """Accessibility utilities and UI element discovery."""

    def __init__(self, client: "IosClient") -> None:
        """Initialize accessibility frameworks and UI client."""
        self._client = client
        self._client.load_framework("AXRuntime")
        self._client.load_framework("AccessibilityUI")
        self._ui_client = (
            client.symbols.objc_getClass("AXUIClient")
            .objc_call("alloc")
            .objc_call(
                "initWithIdentifier:serviceBundleName:",
                client.cf("AXAuditAXUIClientIdentifier"),
                client.cf("AXAuditAXUIService"),
            )
        )

    @property
    def primary_app(self) -> AXElement:
        """Return the primary app AXElement."""
        if not self.enabled:
            raise RpcAccessibilityTurnedOffError()
        primary_app = self._client.symbols.objc_getClass("AXElement").objc_call("primaryApp")
        if primary_app == 0:
            raise RpcFailedToGetPrimaryAppError()
        return self.axelement(primary_app)

    @property
    def enabled(self) -> bool:
        """Return True if accessibility automation is enabled."""
        return bool(
            self._client.symbols._AXSApplicationAccessibilityEnabled() or self._client.symbols._AXSAutomationEnabled()
        )

    @enabled.setter
    def enabled(self, value: bool) -> None:
        """Enable or disable accessibility automation."""
        self._client.symbols._AXSSetAutomationEnabled(int(value))

    def hide_frame(self) -> None:
        """Hide the accessibility highlight frame."""
        self.draw_frame(0, 0, 0, 0)

    def set_frame_style(self, value: int) -> None:
        """Set the highlight frame style."""
        self._ui_client.objc_call(
            "sendSynchronousMessage:withIdentifier:error:", self._client.cf({"frameStyle": value}), 2, 0
        )

    def draw_frame(self, x: float, y: float, width: float, height: float) -> None:
        """Draw a highlight frame at the given coordinates."""
        rect = {"frame": f" {{{{{x},{y}}}, {{{width},{height}}}}}"}
        self._ui_client.objc_call("sendSynchronousMessage:withIdentifier:error:", self._client.cf(rect), 1, 0)

    def wait_for_element_by_label(
        self,
        label: str,
        auto_scroll: bool = True,
        draw_frame: bool = True,
        timeout: float = 5,
        direction: AXDirection = AXDirection.Next,
        displayed_only: bool = False,
    ) -> AXElement:
        """Wait for an element with the given label to appear."""
        start = time.time()
        while time.time() - start < timeout:
            try:
                return self._get_element_by_label(
                    label,
                    auto_scroll=auto_scroll,
                    draw_frame=draw_frame,
                    direction=direction,
                    displayed_only=displayed_only,
                )
            except ElementNotFoundError:
                pass
            except AttributeError:
                pass
        raise ElementNotFoundError(
            f'failed to find AXElement by label: "{label}" after waiting for {timeout} seconds for it to load'
        )

    def _get_element_by_label(
        self,
        label: str,
        auto_scroll: bool = True,
        draw_frame: bool = True,
        direction: AXDirection = AXDirection.Next,
        displayed_only: bool = False,
    ) -> AXElement:
        """Return an AXElement with the given label."""
        if direction == AXDirection.Next:
            element = self.primary_app.first_element
            elements_list = self.primary_app.elements
        elif direction == AXDirection.Previous:
            element = self.primary_app.last_element
            elements_list = reversed(self.primary_app.elements)
        else:
            raise TypeError(f"bad value for: {direction}")

        if displayed_only:
            for element in elements_list:
                if element.compare_label(label, auto_scroll=False, draw_frame=draw_frame):
                    return element
        else:
            while True:
                if element is None:
                    break

                if element.compare_label(label, auto_scroll=auto_scroll, draw_frame=draw_frame):
                    if draw_frame:
                        self.hide_frame()
                    return element

                element = element.next(direction=direction)

        if draw_frame:
            self.hide_frame()

        raise ElementNotFoundError(f'failed to find AXElement by label: "{label}"')

    def press_elements_by_labels(
        self,
        labels: list[str],
        auto_scroll: bool = True,
        draw_frame: bool = True,
        timeout: float = 5,
        direction: AXDirection = AXDirection.Next,
        displayed_only: bool = False,
    ) -> None:
        """
        Press a sequence of labels in order.

        :param labels: Label list to press.
        :param auto_scroll: Scroll to each chosen element.
        :param draw_frame: Draw a frame over the current element.
        :param timeout: Timeout to wait for each element to appear.
        :param direction: Direction to search.
        :param displayed_only: Search only displayed elements.
        """
        for label in labels:
            self.wait_for_element_by_label(
                label,
                auto_scroll=auto_scroll,
                draw_frame=draw_frame,
                timeout=timeout,
                direction=direction,
                displayed_only=displayed_only,
            ).press()

            if draw_frame:
                self.hide_frame()

    def axelement(self, symbol: DarwinSymbol) -> AXElement:
        """Wrap a DarwinSymbol as an AXElement."""
        return AXElement.create(symbol, self._client)



from collections import namedtuple
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import ObjectiveCLexer

from rpcclient.clients.darwin.objective_c import objc
from rpcclient.core.symbols_jar import SymbolsJar
from rpcclient.exceptions import GettingObjectiveCClassError

if TYPE_CHECKING:
    from rpcclient.clients.darwin.client import DarwinClient

Ivar = namedtuple("Ivar", "name type_ offset")


class Class:
    """
    Wrapper for ObjectiveC Class object.
    """

    def __init__(self, client: "DarwinClient", class_object=0, class_data: Optional[dict] = None, lazy=False):
        """
        :param rpcclient.darwin.client.DarwinClient client: Darwin client.
        :param rpcclient.darwin.objective_c_symbol.Symbol class_object:
        """
        self._client = client
        self._class_object = class_object
        self.protocols = []
        self.ivars = []
        self.properties = []
        self.methods = []
        self.name = ""
        self.super = None
        if not lazy:
            if class_data is None:
                self.reload()
            else:
                self._load_class_data(class_data)

    @staticmethod
    def from_class_name(client: "DarwinClient", class_name: str):
        """
        Create ObjectiveC Class from given class name.
        :param rpcclient.darwin.client.DarwinClient client: Darwin client.
        :param class_name: Class name.
        """
        class_object = client.symbols.objc_getClass(class_name)
        if not class_object:
            raise GettingObjectiveCClassError()
        class_symbol = Class(client, class_object)
        if class_symbol.name != class_name:
            raise GettingObjectiveCClassError()
        return class_symbol

    @staticmethod
    def sanitize_name(name: str):
        """
        Sanitize python name to ObjectiveC name.
        """
        name = "_" + name[1:].replace("_", ":") if name.startswith("_") else name.replace("_", ":")
        return name

    def reload(self):
        """
        Reload class object data.
        Should be used whenever the class layout changes (for example, during method swizzling)
        """
        objc_class = self._class_object if self._class_object else self._client.symbols.objc_getClass(self.name)
        class_description = self._client.showclass(objc_class)

        self.super = Class(self._client, class_description["super"]) if class_description["super"] else None
        self.name = class_description["name"]
        self.protocols = class_description["protocols"]
        self.ivars = [
            Ivar(name=ivar["name"], type_=ivar["type"], offset=ivar["offset"]) for ivar in class_description["ivars"]
        ]
        self.properties = [
            objc.Property(name=prop["name"], attributes=objc.convert_encoded_property_attributes(prop["attributes"]))
            for prop in class_description["properties"]
        ]
        self.methods = [objc.Method.from_data(method, self._client) for method in class_description["methods"]]

    def show(self, dump_to: Optional[str] = None):
        """
        Print to terminal the highlighted class description.
        :param dump_to: directory to dump.
        """
        formatted = str(self)
        print(highlight(formatted, ObjectiveCLexer(), TerminalTrueColorFormatter(style="native")))

        if dump_to is None:
            return
        (Path(dump_to) / f"{self.name}.m").expanduser().write_text(formatted)

    def objc_call(self, sel: str, *args, **kwargs):
        """
        Invoke a selector on the given class object.
        :param sel: Selector name.
        :return: whatever the selector returned as a symbol.
        """
        return self._class_object.objc_call(sel, *args, **kwargs)

    def get_method(self, name: str):
        """
        Get a specific method implementation.
        :param name: Method name.
        :return: Method.
        """
        for method in self.methods:
            if method.name == name:
                return method

    def iter_supers(self):
        """
        Iterate over the super classes of the class.
        """
        sup = self.super
        while sup is not None:
            yield sup
            sup = sup.super

    def _load_class_data(self, data: dict):
        self._class_object = self._client.symbol(data["address"])
        self.super = Class(self._client, data["super"]) if data["super"] else None
        self.name = data["name"]
        self.protocols = data["protocols"]
        self.ivars = [Ivar(name=ivar["name"], type_=ivar["type"], offset=ivar["offset"]) for ivar in data["ivars"]]
        self.properties = data["properties"]
        self.methods = data["methods"]

    @property
    def symbols_jar(self) -> SymbolsJar:
        """Get a SymbolsJar object for quick operations on all methods"""
        jar = SymbolsJar.create(self._client)

        for m in self.methods:
            jar[f"[{self.name} {m.name}]"] = m.address

        return jar

    @property
    def bundle_path(self) -> Path:
        return Path(
            self._client.symbols.objc_getClass("NSBundle")
            .objc_call("bundleForClass:", self._class_object)
            .objc_call("bundlePath")
            .py()
        )

    def __dir__(self):
        result = set()

        for method in self.methods:
            if method.is_class:
                result.add(method.name.replace(":", "_"))

        for sup in self.iter_supers():
            for method in sup.methods:
                if method.is_class:
                    result.add(method.name.replace(":", "_"))

        result.update(list(super().__dir__()))
        return list(result)

    def __str__(self):
        protocol_buf = f"<{','.join(self.protocols)}>" if self.protocols else ""

        if self.super is not None:
            buf = f"@interface {self.name}: {self.super.name} {protocol_buf}\n"
        else:
            buf = f"@interface {self.name} {protocol_buf}\n"

        # Add ivars
        buf += "{\n"
        for ivar in self.ivars:
            buf += f"\t{ivar.type_} {ivar.name}; // 0x{ivar.offset:x}\n"
        buf += "}\n"

        # Add properties
        for prop in self.properties:
            buf += f"@property ({','.join(prop.attributes.list)}) {prop.attributes.type_} {prop.name};\n"

            if prop.attributes.synthesize is not None:
                buf += f"@synthesize {prop.name} = {prop.attributes.synthesize};\n"

        # Add methods
        for method in self.methods:
            buf += str(method)

        buf += "@end"
        return buf

    def __repr__(self):
        return f'<objC Class "{self.name}">'

    def __getitem__(self, item):
        for method in self.methods:
            if method.name == item:
                if method.is_class:
                    return partial(self.objc_call, item)
                else:
                    raise AttributeError(f"{self.name} class has an instance method named {item}, not a class method")

        for sup in self.iter_supers():
            for method in sup.methods:
                if method.name == item:
                    if method.is_class:
                        return partial(self.objc_call, item)
                    else:
                        raise AttributeError(
                            f"{self.name} class has an instance method named {item}, not a class method"
                        )

        raise AttributeError(f"""'{self.name}' class has no attribute {item}""")

    def __getattr__(self, item: str):
        return self[self.sanitize_name(item)]



import plistlib
import struct

import lief
from parameter_decorators import path_to_str

from rpcclient.clients.darwin.consts import kSecCodeMagicEntitlement
from rpcclient.core.subsystems.lief import Lief
from rpcclient.exceptions import NoEntitlementsError


class DarwinLief(Lief):
    @path_to_str("path")
    def get_entitlements(self, path: str) -> dict:
        with self._client.fs.open(path, "r") as f:
            buf = f.read()
        parsed = lief.parse(buf)
        code_signature = buf[
            parsed.code_signature.data_offset : parsed.code_signature.data_offset + parsed.code_signature.data_size
        ]

        ent_magic = struct.pack(">I", kSecCodeMagicEntitlement)
        ent_magic_offset = code_signature.find(ent_magic)

        if ent_magic_offset == -1:
            raise NoEntitlementsError()

        ent_buf = code_signature[ent_magic_offset + len(ent_magic) + 4 :]
        end_plist_magic = b"</plist>"
        ent_buf = ent_buf[: ent_buf.find(end_plist_magic) + len(end_plist_magic)]
        return plistlib.loads(ent_buf)



import pytest

from rpcclient.core.subsystems.decorator import SubsystemNotAvailable, subsystem
from rpcclient.exceptions import ArgumentError


def _get_subsystems(client) -> list[str]:
    names: set[str] = set()
    for cls in client.__class__.mro():
        for name, attr in vars(cls).items():
            if isinstance(attr, subsystem):
                names.add(name)
    return sorted(names)


def test_peek(client):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_malloc(0x100) as peekable:
        client.peek(peekable, 0x100)


def test_poke(client):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_malloc(0x100) as peekable:
        client.poke(peekable, b"a" * 0x100)


@pytest.mark.parametrize(
    "params", [([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]), ([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15])]
)
def test_16args(client, params):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_malloc(0x100) as peekable:
        client.symbols.test_16args(peekable, *params)
        for i in range(len(params)):
            assert peekable[i] == params[i]


@pytest.mark.parametrize(
    "va_list_index,params,expected",
    [
        (2, ["%f", 15.5], "15.500000"),
        (2, ["%f %f", 15.5, 7.3], "15.500000 7.300000"),
        (2, ["%f %f %d", 15.5, 7.3, 42], "15.500000 7.300000 42"),
        (2, ["%f %f %d %d", 15.5, 7.3, 1, 2], "15.500000 7.300000 1 2"),
        (2, ["%f %f %d %d %s", 15.5, 7.3, 1, 2, "test"], "15.500000 7.300000 1 2 test"),
        (2, ["%f %f %d %d %s %s", 15.5, 7.3, 1, 2, "test", "test2"], "15.500000 7.300000 1 2 test test2"),
    ],
)
@pytest.mark.arm
def test_va_list_call(client, va_list_index, params, expected):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_malloc(0x100) as peekable:
        client.symbols.sprintf(peekable, *params, va_list_index=va_list_index)
        assert expected == peekable.peek_str()


@pytest.mark.arm
def test_floating_point_call(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.symbols.sqrt(16.0, return_float64=True) == 4.0


def test_peek_invalid_address(client):
    """
    :param rpcclient.client.Client client:
    """
    # Server config: -DSAFE_WRITE raises ArgumentError for invalid address; else, ConnectionError.
    with pytest.raises((ArgumentError, ConnectionError)):
        client.peek(0, 0x10)


def test_poke_invalid_address(client):
    """
    :param rpcclient.client.Client client:
    """
    # Server config: -DSAFE_WRITE raises ArgumentError for invalid address; else, ConnectionError.
    with pytest.raises((ArgumentError, ConnectionError)):
        client.poke(0, b"a")


def test_get_dummy_block(client):
    """
    :param rpcclient.client.Client client:
    """
    client.cf([1, 2, 3]).objc_call("enumerateObjectsUsingBlock:", client.get_dummy_block())


def test_listdir(client):
    """
    :param rpcclient.client.Client client:
    """
    entries = client.listdir("/")
    assert entries[0].d_name == "."
    assert entries[1].d_name == ".."


def test_calloc(client):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_calloc(0x100) as zeros:
        assert client.peek(zeros, 0x100) == b"\x00" * 0x100


def test_env_get_set(client):
    """
    :param rpcclient.client.Client client:
    """
    client.setenv("TEST", "test")
    assert client.getenv("TEST") == "test"


def test_environ(client):
    """
    :param rpcclient.client.Client client:
    """
    assert len(client.environ) > 0
    for e in client.environ:
        assert "=" in e


def test_all_subsystems_initialize(client):
    """
    :param rpcclient.client.Client client:

    Ensure each @subsystem property is initialized and not SubsystemNotAvailable.
    """
    subsystem_names = _get_subsystems(client)
    failures: dict[str, str] = {}

    for name in subsystem_names:
        value = getattr(client, name)
        if isinstance(value, SubsystemNotAvailable):
            failures[name] = repr(value)

    assert not failures, f"Subsystems failed to initialize: {failures}"



from parameter_decorators import path_to_str

from rpcclient.clients.darwin.structs import stat64, statfs64
from rpcclient.core.subsystems.fs import Fs, RemotePath


def do_stat(client, stat_name, filename: str):
    """Return a stat64 struct for a remote path."""
    with client.safe_malloc(stat64.sizeof()) as buf:
        err = client.symbols[stat_name](filename, buf)
        if err != 0:
            client.raise_errno_exception(f"failed to stat(): {filename}")
        return stat64.parse_stream(buf)


class DarwinRemotePath(RemotePath):
    def __init__(self, path: str, client) -> None:
        super().__init__(path, client)

    def stat(self):
        return do_stat(self._client, "stat64", self._path)

    def lstat(self):
        return do_stat(self._client, "lstat64", self._path)


class DarwinFs(Fs):
    @path_to_str("path")
    def stat(self, path: str):
        """Return stat64 info for a remote path."""
        return do_stat(self._client, "stat64", path)

    @path_to_str("path")
    def lstat(self, path: str):
        """Return lstat64 info for a remote path (does not follow symlinks)."""
        return do_stat(self._client, "lstat64", path)

    @path_to_str("path")
    def setxattr(self, path: str, name: str, value: bytes):
        """set an extended attribute value"""
        count = self._client.symbols.setxattr(path, name, value, len(value), 0, 0).c_int64
        if count == -1:
            self._client.raise_errno_exception(f"failed to setxattr(): {path}")

    @path_to_str("path")
    def removexattr(self, path: str, name: str):
        """remove an extended attribute value"""
        count = self._client.symbols.removexattr(path, name, 0).c_int64
        if count == -1:
            self._client.raise_errno_exception(f"failed to removexattr(): {path}")

    @path_to_str("path")
    def listxattr(self, path: str) -> list[str]:
        """list extended attribute names"""
        max_buf_len = 1024
        with self._client.safe_malloc(max_buf_len) as xattributes_names:
            count = self._client.symbols.listxattr(path, xattributes_names, max_buf_len, 0).c_int64
            if count == -1:
                self._client.raise_errno_exception(f"failed to listxattr(): {path}")
            return [s.decode() for s in xattributes_names.peek(count).split(b"\x00")[:-1]]

    @path_to_str("path")
    def getxattr(self, path: str, name: str) -> bytes:
        """get an extended attribute value"""
        max_buf_len = 1024
        with self._client.safe_malloc(max_buf_len) as value:
            count = self._client.symbols.getxattr(path, name, value, max_buf_len, 0, 0).c_int64
            if count == -1:
                self._client.raise_errno_exception(f"failed to getxattr(): {path}")
            return value.peek(count)

    @path_to_str("path")
    def dictxattr(self, path: str) -> dict[str, bytes]:
        """get a dictionary of all extended attributes"""
        result = {}
        for k in self.listxattr(path):
            result[k] = self.getxattr(path, k)
        return result

    @path_to_str("path")
    def statfs(self, path: str):
        with self._client.safe_malloc(statfs64.sizeof()) as buf:
            if self._client.symbols.statfs64(path, buf) != 0:
                self._client.raise_errno_exception(f"statfs failed for: {path}")
            return statfs64.parse_stream(buf)

    @path_to_str("path")
    def chflags(self, path: str, flags: int) -> None:
        """Set BSD file flags on a remote path."""
        if self._client.symbols.chflags(path, flags) != 0:
            self._client.raise_errno_exception(f"chflags failed for: {path}")

    def remote_path(self, path: str) -> DarwinRemotePath:
        return DarwinRemotePath(path, self._client)



from typing import Any

import click
import inquirer3
from inquirer3.themes import GreenPassion


def prompt_selection(choices: list[Any], message: str, idx: bool = False) -> Any:
    """
    Prompt the user to select a value from a list.

    - choices: iterable of options to present.
    - message: prompt message shown to the user.
    - idx: when True, return the index of the selected item; otherwise return the item itself.

    Raises:
        click.ClickException: if the user cancels the prompt (Ctrl-C).
    """
    question = [inquirer3.List("selection", message=message, choices=choices, carousel=True)]
    try:
        result = inquirer3.prompt(question, theme=GreenPassion(), raise_keyboard_interrupt=True)
    except KeyboardInterrupt as e:
        raise click.ClickException("No selection was made") from e
    return result["selection"] if not idx else choices.index(result["selection"])
